# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
import logging
import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type

import torch
import torchmetrics
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.rank_zero import rank_zero_only
from syne_tune import Reporter
from torch.nn import Parameter
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import Dataset

from renate import defaults
from renate.utils.deepspeed import convert_zero_checkpoint_to_fp32_state_dict
from renate.utils.distributed_strategies import create_strategy
from renate.utils.file import unlink_file_or_folder
from renate.utils.misc import AdditionalTrainingMetrics, int_or_str
from .learner import Learner, ReplayLearner
from ..models import RenateModule


logging_logger = logging.getLogger(__name__)


class SyneTuneCallback(Callback):
    """Callback to report metrics to Syne Tune.

    Args:
        val_enabled: Whether validation was enabled in the Learner.
    """

    def __init__(self, val_enabled: bool):
        super().__init__()
        self._report = Reporter()
        self._val_enabled = val_enabled
        self._additional_metrics = AdditionalTrainingMetrics()

    @rank_zero_only
    def _log(self, trainer: Trainer, pl_module: LightningModule) -> None:
        """Report the current epoch's results to Syne Tune.

        If validation was run `_val_enabled` is True, the results are reported at the end of
        the validation epoch. Otherwise, they are reported at the end of the training epoch.
        """

        training = pl_module.training
        if trainer.sanity_checking or (training and self._val_enabled):
            return
        to_report = {k: v.item() for k, v in trainer.logged_metrics.items()}
        to_report.update(self._additional_metrics(pl_module))
        self._report(
            **to_report,
            step=trainer.current_epoch,
            epoch=trainer.current_epoch + 1,
        )

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._additional_metrics.on_train_epoch_end()
        self._log(trainer=trainer, pl_module=pl_module)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._log(trainer=trainer, pl_module=pl_module)

    def on_train_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self._additional_metrics.on_train_start()


class RenateModelCheckpoint(ModelCheckpoint):
    """Callback to save Renate state after each epoch.

    Args:
        model: Model to be saved when creating a checkpoint.
        output_state_folder: Checkpoint folder location.
        val_enabled: Whether validation was enabled in the Learner. Forwarded to `SyneTuneCallback`.
        metric: Monitored metric to decide when to write a new checkpoint. If no metric is provided
            or validation is not enabled, the latest model will be stored.
        mode: `min` or `max`. Whether to minimize or maximize the monitored `metric`.
        use_syne_tune_callback: Whether to use `SyneTuneCallback`.
    """

    def __init__(
        self,
        model: RenateModule,
        output_state_folder: str,
        val_enabled: bool,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        use_syne_tune_callback: bool = True,
    ) -> None:
        every_n_epochs = 1
        save_last = False
        if metric is None or not val_enabled:
            every_n_epochs = 0
            save_last = True
        learner_checkpoint_filename = Path(defaults.learner_state_file("")).stem
        super().__init__(
            dirpath=output_state_folder,
            filename=learner_checkpoint_filename,
            every_n_epochs=every_n_epochs,
            monitor=metric,
            mode=mode,
            save_last=save_last,
            save_weights_only=True,
        )
        self._model = model
        self._output_state_folder = output_state_folder
        self.CHECKPOINT_NAME_LAST = learner_checkpoint_filename
        # Delete old checkpoint if exists
        unlink_file_or_folder(Path(defaults.learner_state_file(self._output_state_folder)))
        # FIXME: Hack to make sure Syne Tune is called after checkpointing.
        # Details: https://github.com/Lightning-AI/lightning/issues/15026
        # If fixed, remove on_train_epoch_end, on_validation_epoch_end, val_enabled, remove line
        # below, and add in ModelUpdaterSyneTune callback.
        if use_syne_tune_callback:
            self._syne_tune_callback = SyneTuneCallback(val_enabled)
        else:
            self._syne_tune_callback = None

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_train_epoch_end(trainer=trainer, pl_module=pl_module)
        if self._syne_tune_callback is not None:
            self._syne_tune_callback.on_train_epoch_end(trainer=trainer, pl_module=pl_module)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_validation_epoch_end(trainer=trainer, pl_module=pl_module)
        if self._syne_tune_callback is not None:
            self._syne_tune_callback.on_validation_epoch_end(trainer=trainer, pl_module=pl_module)

    def _load_best_checkpoint_and_save(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # Reload best state.
        learner_state_path = Path(defaults.learner_state_file(self._output_state_folder))
        if learner_state_path.exists():
            # There are three obvious steps that are handled by lightning if
            # we use the load_from_checkpoint mechanism. Here we do those manually.
            # See for reference
            # https://github.com/Lightning-AI/lightning/blob/1.8.6/src/pytorch_lightning/core/saving.py#L225
            # 1. Load the state_dict from the checkpoint file.
            # 2. Call the on_load_checkpoint (which is a callback)
            # 3. Load the state_dict into the model. Note the strategy.load_model_state_dict call.
            loaded_state = trainer.strategy.load_checkpoint(learner_state_path)
            pl_module.on_load_checkpoint(loaded_state)
            trainer.strategy.load_model_state_dict(loaded_state)

        # Finalize model update.
        pl_module.on_model_update_end()
        # Save permanently.
        if trainer.is_global_zero:
            # Save the buffer only on rank zero.
            pl_module.save(self._output_state_folder)
        # Overwrite checkpoint.
        self._save_checkpoint(trainer, str(learner_state_path))

    def teardown(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        """Implements the separation of learner and model at the end of training.

        There are two cases two handle.

        1. If deepspeed is being used:
        The learner_state_path (which the checkpointing func) uses is a directory and not a file.
        This directory has sharded state_dicts (of model and optimizers), depending on which
        deepspeed stage is used. There are three steps here

        a. combine all the shards into one big state dict.
        b. The learner_state_path is a dir (learner.ckpt/). This needs to be deleted first.
        c. Write the combined state_dict as the learner.ckpt file as a single file.
        d. Extract the state_dict element from the learner and save that as the model.ckpt.

        2. If not deepspeed (say DDP or single device):
        The steps are much simpler.

        a. Load the learner.ckpt and extract the state_dict element.
        b. | Sanitize the extracted state_dict. Learner has the model in a _model attribute.
           | So strip the first "_model." from the keys of the state_dict.
        c. Save the sanitized model to model.ckpt.

        Case 2 is needs to be done even for Case 1 (step d). So teardown is a recursive call in
        Case 1 which automatically goes to Case 2 as learner.ckpt is file now.
        """

        if trainer.is_global_zero and (stage == "fit"):
            learner_state_path = Path(defaults.learner_state_file(self._output_state_folder))
            if learner_state_path.exists() and learner_state_path.is_dir():
                # Deepspeed zero saves everything as folders.
                combined_state_dict = convert_zero_checkpoint_to_fp32_state_dict(learner_state_path)
                unlink_file_or_folder(learner_state_path)
                torch.save(combined_state_dict, learner_state_path)
                self.teardown(trainer, pl_module, stage)
            elif learner_state_path.exists() and learner_state_path.is_file():
                # This a normal file. We strip the model of any wrappers and save that.
                learner_state = torch.load(learner_state_path)
                out_sd = {
                    k.replace("_model.", "", 1): v for k, v in learner_state["state_dict"].items()
                }  # Replace only 1 instance because we have to load it into RenateModule.
                torch.save(out_sd, defaults.model_file(self.dirpath))
                # Remove model from learner checkpoint
                learner_state["state_dict"] = {}
                torch.save(learner_state, learner_state_path)

    def on_exception(
        self, trainer: Trainer, pl_module: LightningModule, exception: BaseException
    ) -> None:
        super().on_exception(trainer, pl_module, exception)
        self._load_best_checkpoint_and_save(trainer, pl_module)

    def on_fit_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        super().on_fit_end(trainer, pl_module)
        self._load_best_checkpoint_and_save(trainer, pl_module)


class ModelUpdater(abc.ABC):
    """Updates a learner using the data provided.

    Args:
        model: The potentially pretrained model to be updated with new data.
        learner_class: Class of the learner to be used for model update.
        learner_kwargs: Arguments either used for creating a new learner (no previous
            state available) or replace current arguments of the learner.
        input_state_folder: Folder used by Renate to store files for current state.
        output_state_folder: Folder used by Renate to store files for next state.
        max_epochs: The maximum number of epochs used to train the model. For comparability between
            methods, epochs are interpreted as "finetuning-equivalent". That is, one epoch is
            defined as `len(current_task_dataset) / batch_size` update steps.
        train_transform: The transformation applied during training.
        train_target_transform: The target transformation applied during testing.
        test_transform: The transformation at test time.
        test_target_transform: The target transformation at test time.
        buffer_transform: Augmentations applied to the input data coming from the memory. Not all
            updaters require this. If required but not passed, `transform` will be used.
        buffer_target_transform: Transformations applied to the target. Not all updaters require
            this. If required but not passed, `target_transform` will be used.
        metric: Monitored metric to decide when to write a new checkpoint or early-stop the
            optimization. If no metric is provided, the latest model will be stored.
        mode: `min` or `max`. Whether to minimize or maximize the monitored `metric`.
        logged_metrics: Metrics logged additional to the default ones.
        early_stopping_enabled: Enables the early stopping of the optimization.
        logger: Logger used by PyTorch Lightning to log intermediate results.
        accelerator: Accelerator used by PyTorch Lightning to train the model.
        devices: Devices used by PyTorch Lightning to train the model. If the devices flag is not
            defined, it will assume devices to be "auto" and fetch the `auto_device_count` from the
            `accelerator`.
        deterministic_trainer: When set to True makes the output of the training deterministic.
            The value is passed to the trainer as described
            `here <https://pytorch-lightning.readthedocs.io/en/stable/common\
            /trainer.html#reproducibility>`_.
        gradient_clip_val: Gradient clipping value used in PyTorch Lightning. Defaults to not
            clipping by using a value of None.
        gradient_clip_algorithm: Method to clip gradients (norm or value) used in PyTorch Lightning.
    """

    def __init__(
        self,
        model: RenateModule,
        loss_fn: torch.nn.Module,
        optimizer: Callable[[List[Parameter]], Optimizer],
        learner_class: Type[Learner],
        learner_kwargs: Optional[Dict[str, Any]] = None,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        learning_rate_scheduler: Optional[Optional[Callable[[Optimizer], _LRScheduler]]] = None,
        learning_rate_scheduler_interval: defaults.SUPPORTED_LR_SCHEDULER_INTERVAL_TYPE = defaults.LR_SCHEDULER_INTERVAL,  # noqa: E501
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        strategy: Optional[str] = defaults.DISTRIBUTED_STRATEGY,
        precision: str = defaults.PRECISION,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
        gradient_clip_val: Optional[float] = defaults.GRADIENT_CLIP_VAL,
        gradient_clip_algorithm: Optional[str] = defaults.GRADIENT_CLIP_ALGORITHM,
        mask_unused_classes: bool = defaults.MASK_UNUSED_CLASSES,
    ):
        self._learner_kwargs = learner_kwargs or {}
        self._learner_kwargs["loss_fn"] = loss_fn
        self._learner_kwargs["optimizer"] = optimizer
        self._learner_kwargs["mask_unused_classes"] = mask_unused_classes
        if learning_rate_scheduler is not None:
            self._learner_kwargs["learning_rate_scheduler"] = learning_rate_scheduler
            self._learner_kwargs[
                "learning_rate_scheduler_interval"
            ] = learning_rate_scheduler_interval
        self._model = model
        self._learner_state_file: Optional[str] = None
        if input_state_folder is not None:
            self._learner_state_file = defaults.learner_state_file(input_state_folder)
        else:
            logging_logger.info(
                "No location for current updater state provided. Updating will start from scratch."
            )
        if output_state_folder is None:
            logging_logger.info(
                "No location for next updater state provided. No state will be stored."
            )
        elif metric is None:
            logging_logger.info(
                "Metric or mode is not provided. Checkpoint is saved only after training."
            )
        if metric is None and early_stopping_enabled:
            warnings.warn(
                "Early stopping is enabled but no metric is specified. Early stopping will be "
                "ignored."
            )
            early_stopping_enabled = False

        self._input_state_folder = input_state_folder
        self._output_state_folder = output_state_folder
        self._metric = metric
        self._mode = mode
        self._logged_metrics = logged_metrics
        self._early_stopping_enabled = early_stopping_enabled
        self._train_transform = train_transform
        self._train_target_transform = train_target_transform
        self._test_transform = test_transform
        self._test_target_transform = test_target_transform
        self._buffer_transform = buffer_transform or train_transform
        self._buffer_target_transform = buffer_target_transform or train_target_transform
        self._transforms_kwargs = {
            "train_transform": self._train_transform,
            "train_target_transform": self._train_target_transform,
            "test_transform": self._test_transform,
            "test_target_transform": self._test_target_transform,
        }
        if issubclass(learner_class, ReplayLearner):
            self._transforms_kwargs["buffer_transform"] = self._buffer_transform
            self._transforms_kwargs["buffer_target_transform"] = self._buffer_target_transform
        self._max_epochs = max_epochs
        if accelerator not in defaults.SUPPORTED_ACCELERATORS:
            raise ValueError(
                f"Accelerator {accelerator} not supported. "
                f"Supported accelerators are {defaults.SUPPORTED_ACCELERATORS}."
            )
        self._accelerator = accelerator
        self._devices = devices
        self._strategy = strategy
        self._precision = int_or_str(precision)
        self._learner = self._load_learner(learner_class, self._learner_kwargs)
        assert self._learner.is_logged_metric(metric), f"Target metric `{metric}` is not logged."
        self._logger = logger
        self._num_epochs_trained = 0
        self._deterministic_trainer = deterministic_trainer
        self._gradient_clip_algorithm = gradient_clip_algorithm
        self._gradient_clip_val = gradient_clip_val

    @abc.abstractmethod
    def update(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Updates the model using the data passed as input.

        Args:
            train_dataset: The training data.
            val_dataset: The validation data.
            train_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the training data.
            val_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the validation data.
            task_id: The task id.
        """

    def _load_learner(
        self,
        learner_class: Type[Learner],
        learner_kwargs: Dict[str, Any],
    ) -> Learner:
        if self._learner_state_file is None or not Path(self._learner_state_file).is_file():
            logging_logger.warning("No updater state available. Updating from scratch.")
            return learner_class(
                model=self._model,
                **learner_kwargs,
                logged_metrics=self._logged_metrics,
                **self._transforms_kwargs,
            )
        learner = learner_class.load_from_checkpoint(
            self._learner_state_file,
            model=self._model,
            logged_metrics=self._logged_metrics,
            strict=False,
            **self._transforms_kwargs,
            **learner_kwargs,
        )
        learner.load(self._input_state_folder)
        return learner

    def _fit_learner(
        self,
        learner: Learner,
        use_syne_tune_callback: bool = True,
    ) -> None:
        callbacks: List[Callback] = []
        if use_syne_tune_callback:
            callbacks.append(SyneTuneCallback(learner.val_enabled))
        if self._output_state_folder is not None:
            model_checkpoint_callback = RenateModelCheckpoint(
                model=self._model,
                output_state_folder=self._output_state_folder,
                metric=self._metric,
                mode=self._mode,
                val_enabled=learner.val_enabled,
                use_syne_tune_callback=use_syne_tune_callback,
            )
            callbacks = [model_checkpoint_callback]  # FIXME: insert at 0 as soon as PTL is fixed.

        if self._early_stopping_enabled:
            if learner.val_enabled:
                callbacks.insert(0, EarlyStopping(monitor=self._metric, mode=self._mode))
            else:
                warnings.warn(
                    "Early stopping is currently not supported without a validation set. It will "
                    "be ignored."
                )

        strategy = create_strategy(self._devices, self._strategy)
        # Finetuning-equivalent epochs.
        num_batches = len(learner._train_dataset) // learner._batch_size
        num_batches += min(len(learner._train_dataset) % learner._batch_size, 1)
        trainer = Trainer(
            accelerator=self._accelerator,
            devices=self._devices,
            max_epochs=self._max_epochs,
            limit_train_batches=num_batches,
            callbacks=callbacks,
            logger=self._logger,
            enable_progress_bar=False,
            deterministic=self._deterministic_trainer,
            strategy=strategy,
            precision=self._precision,
            gradient_clip_val=self._gradient_clip_val,
            gradient_clip_algorithm=self._gradient_clip_algorithm,
        )
        trainer.fit(learner)
        self._num_epochs_trained = trainer.current_epoch


class SingleTrainingLoopUpdater(ModelUpdater):
    """Simple ModelUpdater which requires a single learner only to update the model."""

    def update(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        train_dataset_collate_fn: Optional[Callable] = None,
        val_dataset_collate_fn: Optional[Callable] = None,
        task_id: Optional[str] = None,
    ) -> RenateModule:
        """Updates the model using the data passed as input.

        Args:
            train_dataset: The training data.
            val_dataset: The validation data.
            train_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the training data.
            val_dataset_collate_fn: collate_fn used to merge a list of samples to form a
                mini-batch of Tensors for the validation data.
            task_id: The task id.
        """
        self._learner.on_model_update_start(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            train_dataset_collate_fn=train_dataset_collate_fn,
            val_dataset_collate_fn=val_dataset_collate_fn,
            task_id=task_id,
        )
        self._fit_learner(self._learner)
        return self._model



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from torch.utils.data import ConcatDataset, Dataset, Subset
from torchvision.transforms import Lambda, RandomRotation, ToPILImage

from renate import defaults
from renate.benchmark.datasets.base import DataIncrementalDataModule
from renate.data.data_module import RenateDataModule
from renate.data.datasets import _TransformedDataset
from renate.utils.pytorch import get_generator, randomly_split_data


class Scenario(abc.ABC):
    """Creates a continual learning scenario from a RenateDataModule.

    This class can be extended to modify the returned training/validation/test sets
    to implement different experimentation settings.

    Note that many scenarios implemented here perform randomized operations, e.g., to split a base
    dataset into chunks. The scenario is only reproducible if the _same_ seed is provided in
    subsequent instantiations. The seed argument is required for these scenarios.

    Args:
        data_module: The source RenateDataModule for the user data.
        num_tasks: The total number of expected tasks for experimentation.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        num_tasks: int,
        chunk_id: int,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__()
        self._data_module = data_module
        self._num_tasks = num_tasks
        self._verify_chunk_id(chunk_id)
        self._chunk_id = chunk_id
        self._seed = seed
        self._train_data: Optional[Dataset] = None
        self._val_data: Optional[Dataset] = None
        self._test_data: Optional[List[Dataset]] = None
        self._train_collate_fn: Optional[Callable] = None
        self._val_collate_fn: Optional[Callable] = None
        self._test_collate_fn: Optional[Callable] = None

    def prepare_data(self) -> None:
        """Downloads datasets."""
        self._data_module.prepare_data()

    @abc.abstractmethod
    def setup(self) -> None:
        """Sets up the scenario."""
        self._data_module.setup()
        self._train_collate_fn = self._data_module.train_collate_fn()
        self._val_collate_fn = self._data_module.val_collate_fn()
        self._test_collate_fn = self._data_module.test_collate_fn()

    def train_data(self) -> Dataset:
        """Returns training dataset with respect to current `chunk_id`."""
        return self._train_data

    def val_data(self) -> Dataset:
        """Returns validation dataset with respect to current `chunk_id`."""
        return self._val_data

    def test_data(self) -> List[Dataset]:
        """Returns the test data with respect to all tasks in `num_tasks`."""
        return self._test_data

    def train_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for train DataLoader."""
        return self._train_collate_fn

    def val_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for validation DataLoader."""
        return self._val_collate_fn

    def test_collate_fn(self) -> Optional[Callable]:
        """Returns collate_fn for test DataLoader."""
        return self._test_collate_fn

    def _verify_chunk_id(self, chunk_id: int) -> None:
        """A helper function to verify that the `chunk_id` is valid."""
        assert 0 <= chunk_id < self._num_tasks

    def _split_and_assign_train_and_val_data(self) -> None:
        """Performs train/val split and assigns the `train_data` and `val_data` attributes."""
        proportions = [1 / self._num_tasks for _ in range(self._num_tasks)]
        train_data = self._data_module.train_data()
        self._train_data = randomly_split_data(train_data, proportions, self._seed)[self._chunk_id]
        if self._data_module.val_data():
            val_data = self._data_module.val_data()
            self._val_data = randomly_split_data(val_data, proportions, self._seed)[self._chunk_id]


class ClassIncrementalScenario(Scenario):
    """A scenario that creates data chunks from data samples with specific classes from a data
    module.

    This class, upon giving a list describing the separation of the dataset separates the dataset
    with respect to classification labels.

    Note that, in order to apply this scenario, the scenario assumes that the data points in the
    data module are organised into tuples of exactly 2 tensors i.e. `(x, y)` where `x` is the input
    and `y` is the class id.

    Args:
        data_module: The source RenateDataModule for the user data.
        chunk_id: The data chunk to load in for the training or validation data.
        groupings: Tuple of tuples, describing the division of the classes for respective tasks.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        chunk_id: int,
        groupings: Tuple[Tuple[int, ...], ...],
    ) -> None:
        super().__init__(data_module, len(groupings), chunk_id)
        self._class_groupings = groupings

    def setup(self) -> None:
        """Make assignments: val/train/test splits."""
        super().setup()
        self._train_data = self._get_task_subset(
            self._data_module.train_data(), chunk_id=self._chunk_id
        )
        if self._data_module.val_data():
            self._val_data = self._get_task_subset(
                self._data_module.val_data(), chunk_id=self._chunk_id
            )
        self._test_data = [
            self._get_task_subset(self._data_module.test_data(), i) for i in range(self._num_tasks)
        ]

    def _get_task_subset(self, dataset: Dataset, chunk_id: int) -> Dataset:
        """A helper function identifying indices corresponding to given classes."""
        class_group = self._class_groupings[chunk_id]
        indices = torch.tensor(
            [i for i in range(len(dataset)) if dataset[i][1] in class_group],
            dtype=torch.long,
        )
        subset = Subset(dataset, indices)
        targets = set(np.unique([subset[i][1].item() for i in range(len(subset))]))
        expected_targets = set(self._class_groupings[chunk_id])
        if targets != expected_targets:
            raise ValueError(
                f"Chunk {chunk_id} does not contain classes "
                f"{sorted(list(expected_targets - targets))}."
            )
        return subset


class TransformScenario(Scenario):
    """A scenario that applies a different transformation to each chunk.

    The base ``data_module`` is split into ``len(transforms)`` random chunks. Then ``transforms[i]``
    is applied to chunk ``i``.

    Args:
        data_module: The base data module.
        transforms: A list of transformations.
        chunk_id: The id of the chunk to retrieve.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        transforms: List[Callable],
        chunk_id: int,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__(data_module, len(transforms), chunk_id, seed)
        self._transforms = transforms

    def setup(self) -> None:
        super().setup()
        self._split_and_assign_train_and_val_data()
        self._train_data = _TransformedDataset(
            self._train_data, transform=self._transforms[self._chunk_id]
        )
        if self._val_data:
            self._val_data = _TransformedDataset(
                self._val_data, transform=self._transforms[self._chunk_id]
            )
        self._test_data = []
        for i in range(self._num_tasks):
            self._test_data.append(
                _TransformedDataset(self._data_module.test_data(), transform=self._transforms[i])
            )


class ImageRotationScenario(TransformScenario):
    """A scenario that rotates the images in the dataset by a different angle for each chunk.

    Args:
        data_module: The base data module.
        degrees: List of degrees corresponding to different tasks.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        degrees: List[int],
        chunk_id: int,
        seed: int,
    ) -> None:
        transforms = [RandomRotation(degrees=(deg, deg)) for deg in degrees]
        super().__init__(data_module, transforms, chunk_id, seed)


class PermutationScenario(TransformScenario):
    """A scenario that applies a different random permutation of features for each chunk.

    Args:
        data_module: The base data module.
        num_tasks: The total number of expected tasks for experimentation.
        input_dim: Dimension of the inputs. Can be a shape tuple or the total number of features.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: A random seed to fix the random number generation for permutations.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        num_tasks: int,
        input_dim: Union[List[int], Tuple[int], int],
        chunk_id: int,
        seed: int,
    ) -> None:
        input_dim = np.prod(input_dim)
        rng = get_generator(seed)
        transforms = [torch.nn.Identity()]
        for _ in range(num_tasks - 1):
            permutation = torch.randperm(input_dim, generator=rng)
            transform = Lambda(lambda x, p=permutation: x.flatten()[p].view(x.size()))
            transforms.append(transform)
        super().__init__(data_module, transforms, chunk_id, seed)


class IIDScenario(Scenario):
    """A scenario splitting datasets into random equally-sized chunks."""

    def setup(self) -> None:
        """Make assignments: val/train/test splits."""
        super().setup()
        proportions = [1 / self._num_tasks for _ in range(self._num_tasks)]
        self._train_data = randomly_split_data(
            self._data_module.train_data(), proportions, self._seed
        )[self._chunk_id]
        val_data = self._data_module.val_data()
        if val_data:
            self._val_data = randomly_split_data(val_data, proportions, self._seed)[self._chunk_id]
        self._test_data = [self._data_module.test_data() for _ in range(self._num_tasks)]


class _SortingScenario(Scenario):
    """A scenario that _softly_ sorts a dataset by some score.

    Randomness in the sorted order is induced by swapping the position of random pairs.

    Args:
        data_module: The source RenateDataModule for the user data.
        num_tasks: The total number of expected tasks for experimentation.
        randomness: A value between 0 and 1. For a dataset with ``N`` data points,
            ``0.5 * N * randomness`` random pairs are swapped.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        num_tasks: int,
        randomness: float,
        chunk_id: int,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__(data_module, num_tasks, chunk_id, seed)
        self._randomness = randomness
        assert 0 <= self._randomness <= 1

    @abc.abstractmethod
    def _get_scores(self, dataset: Dataset) -> List[float]:
        """Returns a float value for each data point which is used for sorting."""
        pass

    def _split(self, dataset: Dataset) -> List[Dataset]:
        """Sorts data, applies random swapping and then returns the Datasets."""
        scores = self._get_scores(dataset)
        idx_ordered = [x for _, x in sorted(zip(scores, np.arange(len(dataset))))]
        rng = np.random.RandomState(seed=self._seed)
        for _ in range(int(self._randomness * len(dataset) / 2)):
            i, j = rng.randint(len(dataset), size=2)
            idx_ordered[i], idx_ordered[j] = idx_ordered[j], idx_ordered[i]
        split = torch.tensor_split(torch.tensor(idx_ordered), self._num_tasks)
        return [Subset(dataset, idx) for idx in split]

    def setup(self) -> None:
        """Make assignments: val/train/test splits."""
        super().setup()
        train_data = self._data_module.train_data()
        self._train_data = self._split(train_data)[self._chunk_id]
        val_data = self._data_module.val_data()
        if val_data:
            self._val_data = self._split(val_data)[self._chunk_id]
        test_data = self._data_module.test_data()
        self._test_data = self._split(test_data)


class FeatureSortingScenario(_SortingScenario):
    """A scenario that _softly_ sorts a dataset by the value of a feature, then creates chunks.

    This scenario sorts the data according to a feature value (see `feature_idx`) and randomly
    swaps data positions based on the degree of randomness (see `randomness`).

    This scenario assumes that `dataset[i]` returns a tuple `(x, y)` with a tensor `x` containing
    the features.

    Args:
        data_module: The source RenateDataModule for the user data.
        num_tasks: The total number of expected tasks for experimentation.
        feature_idx: Index of the feature by which to sort. This index refers to the input features
            `x` of a single data point, i.e., no batch dimension. If the tensor `x` has more than
            one dimension, this indexes along the 0-dim while additional dimensions will be averaged
            out. Hence, for images, `feature_idx` refers to a color channel, and we sort by mean
            color channel value.
        randomness: A value between 0 and 1. For a dataset with ``N`` data points,
            ``0.5 * N * randomness`` random pairs are swapped.
        chunk_id: The data chunk to load in for the training or validation data.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        num_tasks: int,
        feature_idx: int,
        randomness: float,
        chunk_id: int,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__(
            data_module=data_module,
            num_tasks=num_tasks,
            randomness=randomness,
            chunk_id=chunk_id,
            seed=seed,
        )
        self._feature_idx = feature_idx

    def _get_scores(self, dataset: Dataset) -> List[float]:
        return [x[0][self._feature_idx].mean().item() for x, _ in dataset]


class HueShiftScenario(_SortingScenario):
    """A scenario that sorts an image dataset by the hue value, then creates chunks.

    All images are sorted by hue value and divided into ``num_tasks`` tasks.
    ``randomness`` is a value between 0 and 1 and controls the number of random swaps applied
    to the sorting.

    This scenario assumes that `dataset[i]` returns a tuple `(x, y)` with a tensor `x` containing
    an RGB image.
    """

    def _get_scores(self, dataset: Dataset) -> List[float]:
        scores = []
        to_pil_image = ToPILImage()
        for image, _ in dataset:
            count, value = np.histogram(
                np.array(to_pil_image(image).convert("HSV"))[:, :, 0].reshape(-1), bins=100
            )
            scores.append(value[np.argmax(count)])
        return scores


class DataIncrementalScenario(Scenario):
    """Creating a scenario which iterates over pre-defined datasets.

    The scenario will iterate over a list of datasets that are provided by the given ``DataModule``.
    The data is loaded by assigning ``data_ids[chunk_id]`` to the attribute of the ``DataModule``
    with name ``domain`` and then calling its ``setup()`` function.

    Args:
        data_module: The source :py:class:`~renate.data.data_module.RenateDataModule` for the user
            data.
        chunk_id: The data chunk to load in for the training or validation data.
        data_ids: Unique identifier for each pre-defined dataset.
        groupings: Tuple of tuples that group different datasets associated to a ``data_id`` to
            one dataset.
        seed: Seed used to fix random number generation.
    """

    def __init__(
        self,
        data_module: RenateDataModule,
        chunk_id: int,
        data_ids: Optional[Tuple[Union[int, str], ...]] = None,
        groupings: Optional[Tuple[Tuple[int, ...], ...]] = None,
        seed: int = defaults.SEED,
    ) -> None:
        if not isinstance(data_module, DataIncrementalDataModule):
            raise ValueError(
                "This scenario is only compatible with classes that extend "
                "`DataIncrementalDataModule`."
            )
        if data_ids is None and groupings is None:
            raise ValueError(
                "Either `data_ids` or `groupings` must be provided. None was provided."
            )
        if data_ids is not None and groupings is not None:
            raise ValueError(
                "Either `data_ids` or `groupings` must be provided. Both were provided."
            )
        if data_ids is not None:
            self._groupings = tuple((data_id,) for data_id in data_ids)
        else:
            self._groupings = groupings
        super().__init__(
            data_module=data_module, num_tasks=len(self._groupings), chunk_id=chunk_id, seed=seed
        )

    def prepare_data(self) -> None:
        """Downloads datasets."""
        for data_id in set(data_id for grouping in self._groupings for data_id in grouping):
            self._data_module.data_id = data_id
            self._data_module.prepare_data()

    def _create_dataset_from_grouping(self, grouping) -> Tuple[Dataset, Optional[Dataset], Dataset]:
        train_data = []
        val_data = []
        test_data = []
        for data_id in grouping:
            self._data_module.data_id = data_id
            self._data_module.setup()
            train_data.append(self._data_module.train_data())
            if self._data_module.val_data() is not None:
                val_data.append(self._data_module.val_data())
            test_data.append(self._data_module.test_data())
        if len(grouping) == 1:
            return train_data[0], val_data[0] if len(val_data) else None, test_data[0]
        return (
            ConcatDataset(train_data),
            ConcatDataset(val_data) if len(val_data) else None,
            ConcatDataset(test_data),
        )

    def setup(self) -> None:
        """Sets up the scenario."""
        super().setup()
        self._train_data, self._val_data, _ = self._create_dataset_from_grouping(
            self._groupings[self._chunk_id]
        )
        self._test_data = [
            self._create_dataset_from_grouping(grouping)[2] for grouping in self._groupings
        ]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union

import torch
from torch.utils.data import default_collate
from transformers import BatchEncoding, DataCollatorWithPadding


@dataclass
class DataCollatorWithPaddingForWildTime(DataCollatorWithPadding):
    """A data collator class that can handle wild time data (non-standard) batches.

    This adds to the `transformer` library's DataCollatorWithPadding. That data collator expects
    data in a standard HF format. Wild time data format is slightly different: We get a
    tuple of BatchEncoding (dict) and a class label. When being read from a buffer, an additional
    metadata attribute is present. These cases are not handled by the orig data collator.
    The code here only separates the input data into format original collator can handle and
    undoes the data packing: see parts after super().__call__.
    """

    def __call__(self, features: List[Dict[str, Any]]) -> Dict[str, Any]:
        # first_element type determines if data sources
        # known types : {features}, class    (from wildtimedataset)
        #             : ({features}, class), metadata (from a replay buffer : val data?)
        #             : index, ({features}, class) (by _EnumeratedDataset in BaseER traindata)
        # Note that {features} can be a dict or BatchEncoding.
        first_element = features[0][0]
        assert len(features[0]) == 2, "Input to collator should contain only two elements"
        if isinstance(first_element, (BatchEncoding, dict)):
            # This is normal dataset and not a buffer with metadata, do the default things
            return self._collate_hf_class_tuples(features)
        elif isinstance(first_element, tuple):
            # this has metadata possibly.
            collated, labels = self._collate_hf_class_tuples([elem[0] for elem in features])
            metadata_to_collate = [elem[1] for elem in features]
            collated_metadata = default_collate(metadata_to_collate)
            return (collated, labels), collated_metadata
        elif isinstance(first_element, (torch.Tensor, int)):
            ## this is EnumeratedDataset.
            collated_indices = default_collate([elem[0] for elem in features])
            collated, labels = self._collate_hf_class_tuples([elem[1] for elem in features])
            return collated_indices, (collated, labels)

        else:
            raise ValueError(
                f"Unknown structure to collate. Got {first_element} of {type(first_element)}"
            )

    def _collate_hf_class_tuples(
        self, features: Tuple[Union[BatchEncoding, Dict[str, Any]], Union[torch.Tensor, int]]
    ):
        to_be_collated = []
        for elem in features:
            elem[0]["labels"] = elem[1]
            to_be_collated.append(elem[0])
        collated = super().__call__(to_be_collated)
        labels = collated.pop("labels")
        return collated, labels



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.utils.misc import (
    AdditionalTrainingMetrics,
    int_or_str,
    maybe_populate_mask_and_ignore_logits,
)


@pytest.mark.parametrize(
    "data_type,target",
    [
        ["16", 16],
        ["32", 32],
        ["bfloat", "bfloat"],
        ["bfloat16", "bfloat16"],
        ["notdata", "notdata"],
    ],
)
def test_int_or_str(data_type, target):
    assert int_or_str(data_type) == target


@pytest.mark.parametrize(
    "use_masking, class_mask, classes_in_task, logits, correct_output",
    [
        [False, None, None, 0.5 * torch.ones((1, 5)), (0.5 * torch.ones((1, 5)), None)],
        [
            True,
            None,
            {0, 1, 2},
            0.5 * torch.ones((1, 5)),
            (
                torch.tensor([[0.5000, 0.5000, 0.5000, -float("inf"), -float("inf")]]),
                torch.tensor([3, 4]),
            ),
        ],
        [
            True,
            torch.tensor([3, 4]),
            {0, 1, 2},
            0.5 * torch.ones((1, 5)),
            (
                torch.tensor([[0.5000, 0.5000, 0.5000, -float("inf"), -float("inf")]]),
                torch.tensor([3, 4]),
            ),
        ],
    ],
)
def test_possibly_populate_mask_and_ignore_logits(
    use_masking, class_mask, classes_in_task, logits, correct_output
):
    out_logits, out_cm = maybe_populate_mask_and_ignore_logits(
        use_masking=use_masking,
        class_mask=class_mask,
        classes_in_current_task=classes_in_task,
        logits=logits,
    )

    # There are a few cases to test. Besides the obvious ones, we also check if the function is a
    # no-op for some parameters. For eg: when a class_mask is provided, it should be returned as is
    # All operations on logits should be inplace. Both are accomplished through comparing
    # data_ptr().
    assert torch.equal(out_logits, correct_output[0])
    if out_cm is None:
        assert out_cm == correct_output[1]
    else:
        assert torch.equal(out_cm, correct_output[1])

    assert out_logits.data_ptr() == logits.data_ptr()
    if class_mask is not None:
        assert class_mask.data_ptr() == out_cm.data_ptr()


@pytest.mark.parametrize(
    "model,gnd",
    [
        (
            torch.nn.Linear(2, 2),
            {"peak_memory_usage": 0, "trainable_params": 6, "total_params": 6},
        ),
        (
            torch.nn.Linear(2, 2).requires_grad_(False),
            {"peak_memory_usage": 0, "trainable_params": 0, "total_params": 6},
        ),
    ],
)
def test_addition_metrics(model, gnd):
    metrics = AdditionalTrainingMetrics()
    out = metrics(model)
    for k in gnd:
        assert gnd[k] == out[k]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, Optional, Tuple

import torch
from torchmetrics import Accuracy

from dummy_datasets import DummyTorchVisionDataModule
from renate.benchmark.models.mlp import MultiLayerPerceptron
from renate.benchmark.scenarios import ClassIncrementalScenario, Scenario
from renate.models import RenateModule


def model_fn(model_state_url: Optional[str] = None) -> RenateModule:
    if model_state_url is None:
        return MultiLayerPerceptron(5 * 5, 10, 0, 64)
    state_dict = torch.load(model_state_url)
    return MultiLayerPerceptron.from_state_dict(state_dict)


def data_module_fn(
    data_path: str,
    chunk_id: Optional[int] = None,
    val_size: float = 0.0,
    seed: int = 0,
    groupings: Tuple[Tuple[int, ...], ...] = ((0, 1), (2, 3, 4)),
) -> Scenario:
    data_module = DummyTorchVisionDataModule(transform=None, val_size=val_size, seed=seed)
    return ClassIncrementalScenario(
        data_module=data_module,
        chunk_id=chunk_id,
        groupings=groupings,
    )


def loss_fn(updater: Optional[str] = None) -> torch.nn.Module:
    if updater.startswith("Avalanche-"):
        return torch.nn.CrossEntropyLoss()
    return torch.nn.CrossEntropyLoss(reduction="none")


def metrics_fn() -> Dict:
    return {"accuracy": Accuracy(task="multiclass", num_classes=10)}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import functools
import logging
from typing import Callable, Optional, Union

import torch
import torch.nn as nn

from renate import defaults
from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.benchmark.models.transformer import HuggingFaceSequenceClassificationTransformer
from renate.benchmark.models.vision_transformer import VisionTransformer
from renate.models.prediction_strategies import PredictionStrategy


logger = logging.getLogger(__name__)


class PromptPool(nn.Module):
    """Implements the prompt pool for L2P

    Args:
        pool_size: Total size of the prompt pool.
        pool_selection_size: Number of prompts to select from the pool.
        prompt_size: Number of tokens each prompt is equivalent to.
        prompt_key_dim: Dimensions of the prompt key used to compute similarity. It has to be same
            to the dimensions of `x` in forward.
        embedding_dim: Output dimension of the token/patch embedding layer
        train_prompt_keys: Whether to train the prompt keys. Currently unused.
        similarity_fn: Similarity function between input features and prompt keys
        per_batch_prompt: Flag to use the same prompts for all elements in the batch
    """

    def __init__(
        self,
        pool_size: int = 10,
        pool_selection_size: int = 5,
        prompt_size: int = 5,
        prompt_key_dim: int = 768,
        embedding_dim: int = 768,
        train_prompt_keys: bool = True,
        similarity_fn: Union[Callable, str] = "cosine",
        per_batch_prompt: bool = True,
    ):
        super().__init__()
        self._M = pool_size  ## total pool size
        self._N = pool_selection_size  ## number of prompts selected per input
        self._Lp = prompt_size  ## each prompt is equal to how many tokens
        self._d = embedding_dim  ##
        self._pd = prompt_key_dim
        self._per_batch_prompt = per_batch_prompt

        self._parse_similarity_fn(similarity_fn)
        self.train_prompt_keys = train_prompt_keys  ## This is unused for now
        self.prompt_pool = nn.Parameter(torch.empty((self._M, self._Lp, self._d)).uniform_(-1, 1))
        self.prompt_keys = nn.Parameter(torch.empty((self._M, self._pd)).uniform_(-1, 1))

        self.key_hist = torch.zeros((self._M,), dtype=torch.float32)

    def _parse_similarity_fn(self, similarity_fn: Union[Callable, str]) -> None:
        if callable(similarity_fn):
            self.similarity_fn = similarity_fn
        elif not isinstance(similarity_fn, str):
            raise ValueError(
                "similarity_fn has to be a callable or a string representing similarity metric. "
                "But got {similarity_fn}"
            )
        elif similarity_fn == "cosine":
            normalization_fn = functools.partial(torch.nn.functional.normalize, p=2)
            self.similarity_fn = lambda x, y: normalization_fn(x).matmul(normalization_fn(y).t())
        else:
            raise ValueError(
                f"Currently only cosine similarity is supported, but got {similarity_fn}"
            )

    def forward(self, x: torch.Tensor, manual_prompt_indices: Optional[torch.LongTensor] = None):
        """
        Args:
            x: Image features extracted. It can be [CLS] token or something else of
                dimension B x self.pd..
            manual_prompt_indices: Indices to manually select prompts from pool, instead of
                selecting from
        """
        if manual_prompt_indices is None:
            similarity_matrix = self.similarity_fn(x, self.prompt_keys)
            _, idx = torch.topk(similarity_matrix, k=self._N, dim=1)
            if self._per_batch_prompt:
                prompt_id, id_counts = torch.unique(idx, return_counts=True, sorted=True)
                if prompt_id.shape[0] < self._M:
                    ## The logic for this is taken from the public l2p implementation.
                    temp_pid = torch.full((self._M,), idx.min(), device=prompt_id.device)
                    temp_pid[: prompt_id.shape[0]] = prompt_id
                    prompt_id = temp_pid

                    temp_idc = torch.zeros((self._M,), device=id_counts.device)
                    temp_idc[: id_counts.shape[0]] = id_counts
                    id_counts = temp_idc

                    _, major_idx = torch.topk(id_counts, k=self._N)
                    idx = prompt_id[major_idx].expand(x.shape[0], -1)  # B, top_k
            loss_value = similarity_matrix[:, idx].sum() / (x.shape[0] * x.shape[0])
        else:
            idx = manual_prompt_indices  # should be of size B, top_k
            loss_value = torch.tensor(0.0, device=x.device)

        selected_prompts = self.prompt_pool[idx].flatten(1, 2)
        return selected_prompts, loss_value


class PromptedTransformer(nn.Module):
    """This generic module is the basic prompted transformer. It takes in a model string and creates
    the appropriate transformer (ViT or Text transformer). If no prompts are provided in the forward
    call, image/text features are returned. If a prompt is provided, it is concatenated to the
    embedding layer output and the resultant features are returned.

    Args:
        pretrained_model_name_or_path: A string that denotes which pretrained model from the HF hub
            to use.
        num_outputs: Size of the output.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        pretrained_model_name_or_path="google/vit-base-patch16-224",
        image_size: int = 32,
        patch_size: int = 4,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim: int = 768,
        mlp_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        num_outputs: int = 10,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ) -> None:
        super().__init__()
        if "vit" in pretrained_model_name_or_path:
            self.transformer = VisionTransformer(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                image_size=image_size,
                patch_size=patch_size,
                num_layers=num_layers,
                num_heads=num_heads,
                hidden_dim=hidden_dim,
                mlp_dim=mlp_dim,
                dropout=dropout,
                attention_dropout=attention_dropout,
                num_outputs=num_outputs,
                prediction_strategy=prediction_strategy,
                add_icarl_class_means=add_icarl_class_means,
            )
            self.is_text_transformer = False
        else:
            self.transformer = HuggingFaceSequenceClassificationTransformer(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                num_outputs=num_outputs,
                prediction_strategy=prediction_strategy,
                add_icarl_class_means=add_icarl_class_means,
            )
            for named_param, value in self.transformer.named_parameters():
                if value.shape[0] == self.transformer._backbone.config.vocab_size:
                    self.word_embeddings = self.transformer.get_submodule(
                        named_param.replace(".weight", "")
                    )
                    break

            self.is_text_transformer = True

        self.transformer._tasks_params.clear()
        self.transformer.eval()
        for p in self.transformer.parameters():
            p.requires_grad_(False)

    def forward(
        self, x: torch.Tensor, prompt: Optional[torch.Tensor] = None, cls_feat: bool = True
    ) -> torch.Tensor:
        """
        Args:
            x: Input torch tensor.
            prompt: Prompt tensor. Defaults to None.
            cls_feat: Whether to extract [CLS] token or to return full feature tensor.
                Ignored for text transformer. Defaults to True.
        """
        if prompt is None:
            return (
                self.transformer.get_features(x)
                if self.is_text_transformer
                else self.transformer.get_features(x, cls_feat=cls_feat)
            )
            # text transformers dont support cls_feat.
        elif self.is_text_transformer:
            # The implicit assumption here is that x for text transformers is the input_ids.
            # This simplified forward pass has 4 steps:
            # 1. Get prompts
            # 2. Get embeddings from inputs.
            # 3. Concat prompt and inputs
            # 4. Forward prop inputs_embeds to get the features.
            inputs_embeds = self.word_embeddings(x["input_ids"])
            if prompt.size(0) != inputs_embeds.size(0):
                prompt = prompt.unsqueeze(0).expand(
                    inputs_embeds.size(0), -1, -1
                )  # Expand one prompt to batch size
            inputs_embeds = torch.cat((prompt, inputs_embeds), dim=1)
            return self.transformer.get_features({"inputs_embeds": inputs_embeds})
        else:
            patch_embeddings = self.transformer.get_submodule("_backbone.embeddings")(x)
            if prompt.size(0) != x.size(0):
                prompt = prompt.unsqueeze(0).expand(
                    x.size(0), -1, -1
                )  # Expand one prompt to batch size# Expand one prompt to batch size
            input_concat_prompt = torch.cat([patch_embeddings, prompt], dim=1)

            encoded_features = self.transformer.get_submodule("_backbone.encoder")(
                input_concat_prompt, return_dict=False
            )[0]
            encoded_features = self.transformer.get_submodule("_backbone.layernorm")(
                encoded_features
            )
            return encoded_features[:, 0, :] if cls_feat else encoded_features


class LearningToPromptTransformer(RenateBenchmarkingModule):
    """
    Implements the vision transformer with prompt pool described in
    Wang, Zifeng, et al. "Learning to prompt for continual learning." Proceedings of the
    IEEE/CVF Conference on Computer Vision and Pattern Recognition. 2022.

    Args:
        pretrained_model_name_or_path: A string that denotes which pretrained model from the HF hub
            to use. If provided, it overrides other arguments about architecture.
        image_size: Size of the input image.
        patch_size: Size of the patches.
        num_layers: Number of Encoder layers.
        num_heads: Number of Attention heads.
        hidden_dim: Size of the Encoder's hidden state.
        mlp_dim: Size of the intermediate Multi-layer Perceptron in the Encoder.
        dropout: Dropout probability.
        attention_dropout: Dropout probability for the attention in the Multi-head Attention layer.
        num_outputs: Size of the output.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
        pool_size: Total size of the prompt pool.
        pool_selection_size: Number of prompts to select from the pool.
        prompt_size: Number of tokens each prompt is equivalent to.
        prompt_key_dim: Dimensions of the prompt key used to compute similarity. It has to be same
            to the dimensions of `x` in forward.
        train_prompt_keys: Whether to train the prompt keys. Currently unused.
        similarity_fn: Similarity function between input features and prompt keys.
        per_batch_prompt: Flag to use the same prompts for all elements in the batch.
        prompt_embedding_features: Image feature type used to compute the similarity to prompt keys.
        patch_pooler: Features to feed the classifier.
    """

    def __init__(
        self,
        pretrained_model_name_or_path="google/vit-base-patch16-224",
        image_size: int = 32,
        patch_size: int = 4,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim: int = 768,
        mlp_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        num_outputs: int = 10,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
        pool_size: int = 10,
        pool_selection_size: int = 5,
        prompt_size: int = 5,
        prompt_key_dim: int = 768,
        train_prompt_keys: bool = True,
        similarity_fn: Union[Callable, str] = "cosine",
        per_batch_prompt: bool = True,
        prompt_embedding_features: str = "cls",
        patch_pooler: str = "prompt_mean",
    ) -> None:
        transformer = PromptedTransformer(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            image_size=image_size,
            patch_size=patch_size,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            num_outputs=num_outputs,
            add_icarl_class_means=add_icarl_class_means,
            prediction_strategy=prediction_strategy,
        )
        prompter = PromptPool(
            embedding_dim=transformer.transformer._embedding_size,
            pool_size=pool_size,
            pool_selection_size=pool_selection_size,
            prompt_size=prompt_size,
            prompt_key_dim=prompt_key_dim,
            train_prompt_keys=train_prompt_keys,
            similarity_fn=similarity_fn,
            per_batch_prompt=per_batch_prompt,
        )

        super().__init__(
            embedding_size=transformer.transformer._embedding_size,
            num_outputs=num_outputs,
            constructor_arguments=dict(
                **transformer.transformer._constructor_arguments,
                pool_size=pool_size,
                pool_selection_size=pool_selection_size,
                prompt_size=prompt_size,
                prompt_key_dim=prompt_key_dim,
                train_prompt_keys=train_prompt_keys,
                similarity_fn=similarity_fn,
                per_batch_prompt=per_batch_prompt,
            ),
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )

        self._backbone = nn.ModuleDict({"transformer": transformer, "prompter": prompter})
        self._is_text_transformer = transformer.is_text_transformer
        self.prompt_embedding_features = prompt_embedding_features
        self.patch_pooler = patch_pooler
        self.similarity_score: Optional[torch.Tensor] = None

        assert self.prompt_embedding_features in [
            "cls",
            "mean",
        ], f"Invalid method to extract prompt embedding features. Got {prompt_embedding_features}"

        assert self.patch_pooler in [
            "cls",
            "mean",
            "prompt_mean",
        ], f"Invalid method to extract prompt embedding features. Got {patch_pooler}"

        for p in self._backbone["prompter"].parameters():
            p.requires_grad = True

        # The backbone's forward is monkey-patched to allow the parent class' forward to work
        # without any manual management.
        self._backbone.forward = self.forward_for_monkey_patching

    def forward_for_monkey_patching(
        self, x: torch.Tensor, task_id: str = defaults.TASK_ID
    ) -> torch.Tensor:
        with torch.no_grad():
            prompt_pool_input = self._backbone["transformer"](x, cls_feat=False)
        if not self._is_text_transformer:
            if self.prompt_embedding_features == "cls":
                # retrieve cls token features. This is used in L2P paper.
                prompt_pool_input = prompt_pool_input[:, 0, :]
            elif self.prompt_embedding_features == "mean":
                # compute mean patch features.
                prompt_pool_input = prompt_pool_input[:, 1:, :].mean(1)
            # Compute the prompts to be stacked
        prompts, prompt_similarity = self._backbone["prompter"](prompt_pool_input)
        self.similarity_score = prompt_similarity
        encoded_features = self._backbone["transformer"](x, prompts, cls_feat=False)
        if self._is_text_transformer:
            return encoded_features
        else:
            if self.patch_pooler == "cls":
                seq_cls_token = encoded_features[:, 0, :]
            elif self.patch_pooler == "mean":
                seq_cls_token = encoded_features[:, 1:, :].mean(1)
            elif self.patch_pooler == "prompt_mean":
                num_prompts = prompts.size(1)
                seq_cls_token = encoded_features[:, -num_prompts:, :].mean(1)
            return seq_cls_token



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, List, Optional, Tuple, Union

import torch
from PIL import Image
from torch import Tensor
from torch.utils.data import Dataset

from renate.types import NestedTensors


class ImageDataset(Dataset):
    """Dataset class for image datasets where the images are loaded as raw images.

    Args:
        data: List of data paths to the images.
        labels: Labels of images.
        transform: Transformation or augmentation to perform on the sample.
        target_transform: Transformation or augmentation to perform on the target.
    """

    def __init__(
        self,
        data: List[str],
        labels: List[int],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        assert len(data) == len(labels)
        self.data = data
        self.labels = labels
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor]:
        with open(self.data[idx], "rb") as f:
            image = Image.open(f)
            image.load()

        # In case of black and white images, we should convert them to rgb
        if len(image.size) <= 2:
            image = image.convert("RGB")

        target = torch.tensor(self.labels[idx], dtype=torch.long)

        if self.transform:
            image = self.transform(image)

        if self.target_transform:
            target = self.target_transform(target)
        return image, target.long()


class NestedTensorDataset(Dataset):
    """A dataset of nested tensors.

    Args:
        nested_tensors: A nested tuple/dict structure of tensors. Tensors need to be of equal size
            along the batch dimension.
    """

    def __init__(self, nested_tensors: NestedTensors):
        self._nested_tensors = nested_tensors
        self._length = self._get_len(nested_tensors)

    def _get_len(self, nested_tensors: NestedTensors, expected_length: Optional[int] = None) -> int:
        """Extracts length of the dataset and checks for consistent length."""
        if isinstance(nested_tensors, torch.Tensor):
            length = nested_tensors.size(0)
            assert length == expected_length or expected_length is None
            return length
        elif isinstance(nested_tensors, tuple):
            for t in nested_tensors:
                expected_length = self._get_len(t, expected_length)
            return expected_length
        elif isinstance(nested_tensors, dict):
            for t in nested_tensors.values():
                expected_length = self._get_len(t, expected_length)
            return expected_length
        else:
            raise TypeError(f"Expected nested dict/tuple of tensors, found {type(nested_tensors)}.")

    def __len__(self):
        """Returns the number of data points in the dataset."""
        return self._length

    @staticmethod
    def _get(storage: NestedTensors, idx: int) -> NestedTensors:
        if isinstance(storage, torch.Tensor):
            return storage[idx]
        elif isinstance(storage, tuple):
            return tuple(NestedTensorDataset._get(t, idx) for t in storage)
        elif isinstance(storage, dict):
            return {key: NestedTensorDataset._get(t, idx) for key, t in storage.items()}
        else:
            raise TypeError(f"Expected nested tuple/dict of tensors, found {type(storage)}.")

    def __getitem__(self, idx: int) -> NestedTensors:
        """Read the item stored at index `idx`."""
        return self._get(self._nested_tensors, idx)


class IndexedSubsetDataset(Dataset):
    """A dataset wrapper to keep specified indexes of a dataset element.

    Subset is indexing rows of a (tensor-)dataset, whereas IndexedSubset keeps specified columns.
    It currently handles Datasets whose elements are tuples.

    Args:
        dataset: The dataset to wrap
        indexes_to_keep: An list or tuple of indices that are to be retained.
    """

    def __init__(self, dataset: Dataset, indexes_to_keep: Union[List, Tuple, int]) -> None:
        self.dataset = dataset
        if isinstance(indexes_to_keep, int):
            indexes_to_keep = [indexes_to_keep]
        self.indexes_to_keep = set(indexes_to_keep)

    def __getitem__(self, index) -> Any:
        curr_item = self.dataset[index]
        # Special handling if indexes_to_keep is a single int
        if len(ret_val := [ci for i, ci in enumerate(curr_item) if i in self.indexes_to_keep]) == 1:
            return ret_val[0]
        else:
            return tuple(ret_val)

    def __len__(self):
        return len(self.dataset)


class _TransformedDataset(Dataset):
    """A dataset wrapper that applies transformations.

    This wrapper assumes that the the passed `dataset` returns a pair `(x, y)`.

    Args:
        dataset: The dataset to wrap.
        transform: Transformation to perform on the covariates `x`.
        target_transform: Transformation to perform on the target `y`.
    """

    def __init__(
        self,
        dataset: Dataset,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        super(Dataset, self).__init__()
        self._dataset = dataset
        self._transform = transform
        self._target_transform = target_transform

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        """Returns the transformed data point."""
        data, target = self._dataset[idx]
        if self._transform is not None:
            data = self._transform(data)
        if self._target_transform is not None:
            target = self._target_transform(target)
        return data, target

    def __len__(self) -> int:
        """Returns the number of data points in the dataset."""
        return len(self._dataset)


class _EnumeratedDataset(Dataset):
    """A dataset wrapper that enumerates data points.

    Args:
        dataset: The dataset to wrap.
    """

    def __init__(self, dataset: Dataset) -> None:
        super().__init__()
        self._dataset = dataset

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, Any]:
        """Returns index and data point."""
        return torch.tensor(idx, dtype=torch.long), self._dataset[idx]

    def __len__(self) -> int:
        """Returns the number of data points in the dataset."""
        return len(self._dataset)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import torch
from torch.nn import functional as F
from torch.nn.modules.batchnorm import _BatchNorm


class ContinualNorm(_BatchNorm):
    """Continual Normalization as a replacement for Batch Normalization.

    Pham, Quang, Chenghao Liu, and Steven Hoi.
    "Continual normalization: Rethinking batch normalization for online continual learning."
    International Conference on Learning Representations (2022).

    It combines Group Normalization with respect to a user-defined `num_groups` parameter, the
    number of groups in Group Normalization, followed by Batch Normalization.

    Args:
        num_features: The number of input features in the channel dimension.
        eps:  A value added to the denominator for numerical stability.
        momentum: the value used for the running_mean and running_var computation.
                  Can be set to ``None`` for cumulative moving average.
        affine: Whether learnable affine parameters are going to be used in Batch Normalization.
        track_running_stats:  Whether running stats are tracked in Batch Normalization.
        device: What device to store the parameters.
        dtype: The data type of the learnable parameters.
        num_groups: The number of groups in the Group Normalization.
    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
        device=None,
        dtype=None,
        num_groups: int = 32,
    ):
        super(ContinualNorm, self).__init__(
            num_features, eps, momentum, affine, track_running_stats, device, dtype
        )
        self._num_groups = num_groups

    def forward(self, input: torch.Tensor):
        return F.batch_norm(
            F.group_norm(input, self._num_groups, None, None, self.eps),
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training,
            self.momentum,
            self.eps,
        )

    def extra_repr(self):
        return super().extra_repr() + f", num_groups={self._num_groups}"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import mock

import pytest

from renate.utils.syne_tune import redirect_to_tmp


@pytest.mark.parametrize(
    "uri, expected_uri, raises",
    [
        ("/opt/ml/checkpoints/0/", "/tmp/checkpoints/0/", False),
        ("/home/checkpoints/0/", "/home/checkpoints/0/", True),
    ],
    ids=["redirect-path-to-tmp-dir", "wrong-checkpoint-path-raise-exception"],
)
@mock.patch.dict(os.environ, {"SM_MODEL_DIR": ""})
def test_redirect_to_tmp_on_sagemaker(uri, expected_uri, raises):
    if raises:
        with pytest.raises(AssertionError):
            redirect_to_tmp(uri)
    else:
        assert redirect_to_tmp(uri) == expected_uri


@pytest.mark.parametrize(
    "uri, expected_uri",
    [
        ("/opt/ml/checkpoints/0/", "/opt/ml/checkpoints/0/"),
        ("/home/checkpoints/0/", "/home/checkpoints/0/"),
    ],
)
def test_redirect_to_tmp(uri, expected_uri):
    """This function is not supposed to change anything if not running on SageMaker."""
    assert redirect_to_tmp(uri) == expected_uri



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
import transformers as transformers

from renate.benchmark.datasets.nlp_datasets import MultiTextDataModule


@pytest.mark.skip(reason="This test create problems with the syne-tune redirect test")
def test_multi_data_nlp_small(tmpdir):
    TRAIN_SIZE = 100
    TEST_SIZE = 100

    data = MultiTextDataModule(
        tmpdir,
        train_size=TRAIN_SIZE,
        test_size=TEST_SIZE,
        data_id="ag_news",
        tokenizer=transformers.DistilBertTokenizer.from_pretrained("distilbert-base-uncased"),
        seed=42,
    )

    data.prepare_data()
    data.setup()

    assert len(data.train_data()) == TRAIN_SIZE
    assert len(data.test_data()) == TEST_SIZE

    first_input_agnews = data.train_data()[0][0]["input_ids"]

    data.data_id = "dbpedia_14"
    data.setup()

    tr_data_dbpedia = data.train_data()
    te_data_dbpedia = data.test_data()
    assert len(tr_data_dbpedia) == TRAIN_SIZE
    assert len(te_data_dbpedia) == TEST_SIZE

    first_input_dbpedia = data.train_data()[0][0]["input_ids"]

    assert not torch.all(torch.eq(first_input_dbpedia, first_input_agnews))


@pytest.mark.skip(reason="This test requires downloading and processing four datasets.")
def test_multi_data_nlp_full(tmpdir):
    TRAIN_SIZE = 115000
    TEST_SIZE = 7600

    for d in MultiTextDataModule.domains:
        data = MultiTextDataModule(
            tmpdir,
            train_size=TRAIN_SIZE,
            test_size=TEST_SIZE,
            data_id=d,
            tokenizer=transformers.DistilBertTokenizer.from_pretrained("distilbert-base-uncased"),
        )

        data.prepare_data()
        data.setup()

        tr_data = data.train_data()
        te_data = data.test_data()
        assert len(tr_data) == TRAIN_SIZE
        assert len(te_data) == TEST_SIZE



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import argparse
import json
from pathlib import Path
import os

import boto3
from syne_tune.backend.sagemaker_backend.sagemaker_utils import get_execution_role

from renate.benchmark.experimentation import execute_experiment_job, experiment_config_file


def load_config(scenario_file, model_file, updater_file, dataset_file):
    cs = {}
    for file in [scenario_file, model_file, updater_file, dataset_file]:
        with open(file) as f:
            cs.update(json.load(f))
    return cs


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    for file_name in ["scenario", "model", "updater", "dataset"]:
        parser.add_argument(
            f"--{file_name}-file",
            type=str,
            required=True,
            help=f"File containing the path to the {file_name} JSON.",
        )
    parser.add_argument(
        f"--backend",
        type=str,
        required=True,
        choices=["local", "sagemaker"],
        help="Whether to run locally or on SageMaker.",
    )
    parser.add_argument(
        f"--job-name",
        type=str,
        required=True,
        help="Name of the job.",
    )
    parser.add_argument(
        f"--test-suite",
        type=str,
        required=True,
        choices=["quick", "main"],
        help="Test suite that is run.",
    )
    parser.add_argument(
        f"--seed",
        type=int,
        default=0,
        help="Seed.",
    )
    parser.add_argument(
        f"--max-time",
        type=int,
        default=12 * 3600,
        help="Maximum execution time.",
    )
    parser.add_argument(
        f"--requirements-file",
        type=str,
        required=False,
        help="Path to requirements file",
    )
    args = parser.parse_args()
    config_space = load_config(
        args.scenario_file, args.model_file, args.updater_file, args.dataset_file
    )
    current_folder = Path(os.path.dirname(__file__))
    requirements_file = args.requirements_file
    if not requirements_file:
        requirements_file = current_folder.parent.parent / "requirements.txt"

    if args.backend == "local":
        experiment_outputs_url = (
            Path("tmp")
            / "renate-integration-tests"
            / args.test_suite
            / args.job_name
            / str(args.seed)
        )
        role = None
    else:
        AWS_ACCOUNT_ID = boto3.client("sts").get_caller_identity().get("Account")
        experiment_outputs_url = (
            f"s3://sagemaker-us-west-2-{AWS_ACCOUNT_ID}/renate-integration-tests/"
            f"{args.test_suite}/{args.job_name}/{args.seed}"
        )
        role = get_execution_role()
    execute_experiment_job(
        backend=args.backend,
        config_file=experiment_config_file(),
        config_space=config_space,
        experiment_outputs_url=experiment_outputs_url,
        mode="max",
        metric="val_accuracy",
        num_updates=config_space["num_tasks"],
        role=role,
        instance_type="ml.g4dn.xlarge",
        max_time=args.max_time,
        seed=args.seed,
        job_name=args.job_name[:36],
        devices=1,
        strategy="ddp",
        requirements_file=args.requirements_file,
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import numpy as np
import pandas as pd
import pytest
import torch
import transformers
from torch.utils.data import Dataset, TensorDataset

from renate.benchmark.datasets.nlp_datasets import HuggingFaceTextDataModule
from renate.benchmark.datasets.vision_datasets import (
    CLEARDataModule,
    TinyImageNetDataModule,
    TorchVisionDataModule,
)
from renate.data import CSVDataModule


def test_csv_data_module(tmpdir):
    target_name = "y"

    # Create toy data
    features = np.random.randint(10, size=(10, 4))
    train_data = pd.DataFrame(features, columns=list("ABCD"), dtype="float")
    train_data[target_name] = np.random.randint(10, size=(10, 1))

    features = np.random.randint(10, size=(3, 4))
    test_data = pd.DataFrame(features, columns=list("ABCD"), dtype="float")
    test_data[target_name] = np.random.randint(10, size=(3, 1))

    for stage in ["train", "test"]:
        data = train_data if stage == "train" else test_data
        data_file_tmp = os.path.join(tmpdir, f"{stage}.csv")
        data.to_csv(data_file_tmp, index=False)

    val_size = 0.2
    csv_module = CSVDataModule(
        data_path=tmpdir,
        train_filename="train.csv",
        test_filename="test.csv",
        target_name=target_name,
        val_size=val_size,
    )
    csv_module.prepare_data()
    csv_module.setup()
    train_data = csv_module.train_data()
    val_data = csv_module.val_data()
    test_data = csv_module.test_data()

    assert len(train_data) == round(10 * (1 - val_size))
    assert isinstance(train_data, Dataset)

    assert len(val_data) == round(10 * val_size)
    assert isinstance(val_data, Dataset)

    assert len(test_data) == 3
    assert isinstance(test_data, TensorDataset)


@pytest.mark.slow
@pytest.mark.parametrize(
    "dataset_name,num_tr,num_te,x_shape",
    [
        ("MNIST", 60000, 10000, (1, 28, 28)),
        ("FashionMNIST", 60000, 10000, (1, 28, 28)),
        ("CIFAR10", 50000, 10000, (3, 32, 32)),
        ("CIFAR100", 50000, 10000, (3, 32, 32)),
    ],
)
def test_torchvision_data_module(tmpdir, dataset_name, num_tr, num_te, x_shape):
    """Test loading of torchvision data."""
    val_size = 0.35
    data_module = TorchVisionDataModule(tmpdir, dataset_name=dataset_name, val_size=val_size)
    data_module.prepare_data()
    data_module.setup()
    train_data = data_module.train_data()
    val_data = data_module.val_data()
    test_data = data_module.test_data()
    assert len(train_data) == round(num_tr * (1 - val_size))
    assert isinstance(train_data, Dataset)
    assert len(val_data) == round(num_tr * val_size)
    assert isinstance(val_data, Dataset)
    assert len(test_data) == num_te
    assert isinstance(test_data, Dataset)
    assert train_data[0][0].size() == test_data[0][0].size() == x_shape


@pytest.mark.slow
@pytest.mark.parametrize(
    "dataset_name,chunk_id,num_tr,num_te",
    [
        ("CLEAR10", 0, 3300, 550),
        ("CLEAR100", 0, 9964, 5000),
    ],
)
def test_clear_data_module(tmpdir, dataset_name, chunk_id, num_tr, num_te):
    """Test loading of CLEAR data."""
    val_size = 0.2
    data_module = CLEARDataModule(
        tmpdir, dataset_name=dataset_name, time_step=chunk_id, val_size=val_size
    )
    data_module.prepare_data()
    data_module.setup()
    train_data = data_module.train_data()
    val_data = data_module.val_data()
    test_data = data_module.test_data()
    assert len(train_data) == round(num_tr * (1 - val_size))
    assert isinstance(train_data, Dataset)
    assert len(val_data) == round(num_tr * val_size)
    assert isinstance(val_data, Dataset)
    assert len(test_data) == num_te
    assert isinstance(test_data, Dataset)


@pytest.mark.slow
def test_tiny_imagenet_data_module(tmpdir):
    num_tr = 100000
    num_te = 10000
    val_size = 0.2
    data_module = TinyImageNetDataModule(tmpdir, val_size=val_size)
    data_module.prepare_data()
    data_module.setup()
    train_data = data_module.train_data()
    val_data = data_module.val_data()
    test_data = data_module.test_data()
    assert len(train_data) == round(num_tr * (1 - val_size))
    assert isinstance(train_data, Dataset)
    assert len(val_data) == round(num_tr * val_size)
    assert isinstance(val_data, Dataset)
    assert len(test_data) == num_te
    assert isinstance(test_data, Dataset)
    assert train_data[0][0].size == test_data[0][0].size == (64, 64)


@pytest.mark.slow
@pytest.mark.parametrize(
    "dataset_name,input_column,target_column",
    [
        ("rotten_tomatoes", "text", "label"),
    ],
)
@pytest.mark.parametrize(
    "tokenizer", [transformers.DistilBertTokenizer.from_pretrained("distilbert-base-uncased")]
)
@pytest.mark.parametrize(
    "tokenizer_kwargs",
    [
        None,
        dict(padding="longest", max_length=512, truncation=True),
    ],
)
def test_hugging_face_data_module(
    tmpdir, dataset_name, input_column, target_column, tokenizer, tokenizer_kwargs
):
    data_module = HuggingFaceTextDataModule(
        data_path=tmpdir,
        dataset_name=dataset_name,
        input_column=input_column,
        target_column=target_column,
        tokenizer=tokenizer,
        tokenizer_kwargs=tokenizer_kwargs,
        val_size=0.2,
    )
    data_module.prepare_data()
    data_module.setup()
    train_data = data_module.train_data()
    val_data = data_module.val_data()
    test_data = data_module.test_data()
    assert isinstance(train_data, Dataset)
    assert isinstance(val_data, Dataset)
    assert isinstance(test_data, Dataset)
    for dataset in [train_data, val_data, test_data]:
        inputs, targets = dataset[0]
        assert isinstance(targets, torch.Tensor)
        assert isinstance(inputs, dict)
        assert set(inputs.keys()) == set(["input_ids", "attention_mask"])
        assert isinstance(inputs["input_ids"], torch.Tensor)
        assert isinstance(inputs["attention_mask"], torch.Tensor)


@pytest.mark.slow
@pytest.mark.parametrize("column", ("input", "target"), ids=("input", "target"))
def test_hugging_face_exception_raised_with_wrong_column(tmpdir, column):
    input_column = "text"
    target_column = "label"
    if column == "input":
        input_column = "WRONG_COLUMN"
    elif column == "target":
        target_column = "WRONG_COLUMN"
    data_module = HuggingFaceTextDataModule(
        data_path=tmpdir,
        dataset_name="rotten_tomatoes",
        input_column=input_column,
        target_column=target_column,
        tokenizer=None,
        val_size=0.2,
    )
    if column == "input":
        with pytest.raises(
            ValueError,
            match="Input column 'WRONG_COLUMN' does not exist in rotten_tomatoes. "
            "Available columns: \\['text', 'label'\\].",
        ):
            data_module.prepare_data()
    elif column == "target":
        with pytest.raises(
            ValueError,
            match="Target column 'WRONG_COLUMN' does not exist in rotten_tomatoes. "
            "Available columns: \\['text', 'label'\\].",
        ):
            data_module.prepare_data()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import time
from typing import Dict, Optional, Set, Tuple, Union
from pytorch_lightning import Callback

import torch

from renate.utils.pytorch import complementary_indices


def int_or_str(x: str) -> Union[str, int]:
    """Function to cast to int or str.

    This is used to tackle precision which can be int (16, 32) or str (bf16)
    """
    try:
        return int(x)
    except ValueError:
        return x


def maybe_populate_mask_and_ignore_logits(
    use_masking: bool,
    class_mask: Optional[torch.Tensor],
    classes_in_current_task: Optional[Set[int]],
    logits: torch.Tensor,
):
    """Snippet to compute which logits to ignore after computing the class mask if required."""
    if use_masking:
        if class_mask is None:
            # Now is the time to repopulate the class_mask
            class_mask = torch.tensor(
                complementary_indices(logits.size(1), classes_in_current_task),
                device=logits.device,
                dtype=torch.long,
            )
            # fill the logits with -inf
        logits.index_fill_(1, class_mask.to(logits.device), -float("inf"))

    return logits, class_mask


class AdditionalTrainingMetrics(Callback):
    def __init__(self) -> None:
        self._training_start_time = None
        self._curr_epoch_end_time = None

    def on_train_start(self) -> None:
        self._training_start_time = time.time()

    def on_train_epoch_end(self) -> None:
        self._curr_epoch_end_time = time.time()

    def __call__(self, model: torch.nn.Module) -> Dict[str, Union[float, int]]:
        if all([self._training_start_time, self._curr_epoch_end_time]):
            total_training_time = self._curr_epoch_end_time - self._training_start_time
        else:
            total_training_time = 0.0
        # maximum amount of memory used in training. This might
        # not be the best choice, but the most convenient.
        peak_memory_usage = (
            torch.cuda.memory_stats()["allocated_bytes.all.peak"]
            if torch.cuda.is_available()
            else 0
        )
        trainable_params, total_params = self.parameters_count(model)

        return dict(
            total_training_time=total_training_time,
            peak_memory_usage=peak_memory_usage,
            trainable_params=trainable_params,
            total_params=total_params,
        )

    def parameters_count(self, model: torch.nn.Module) -> Tuple[int, int]:
        trainable_params, total_params = 0, 0
        for param in model.parameters():
            num_params = param.numel()
            total_params += num_params
            if param.requires_grad:
                trainable_params += num_params

        return trainable_params, total_params



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional

from transformers import AutoModelForTextEncoding, PreTrainedModel

from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.models.prediction_strategies import PredictionStrategy


class FeatureExtractorTextTransformer(PreTrainedModel):
    """This is a facade class to extract the correct output from the transformer model."""

    def __init__(self, pretrained_model_name_or_path: str):
        model = AutoModelForTextEncoding.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path
        )
        super().__init__(model.config)
        self._model = model

    def forward(self, x):
        out = self._model(**x, return_dict=True)
        if hasattr(out, "pooler_output"):
            return out.pooler_output
        else:
            return out.last_hidden_state[:, 0]  # 0th element is used for classification.


class HuggingFaceSequenceClassificationTransformer(RenateBenchmarkingModule):
    """RenateBenchmarkingModule which wraps around Hugging Face transformers.

    Args:
        pretrained_model_name_or_path: Hugging Face model id.
        num_outputs: Number of outputs.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        pretrained_model_name_or_path: str,
        num_outputs: int = 10,
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ):
        model = FeatureExtractorTextTransformer(
            pretrained_model_name_or_path=pretrained_model_name_or_path
        )
        constructor_args = dict(pretrained_model_name_or_path=pretrained_model_name_or_path)
        super().__init__(
            embedding_size=model.config.hidden_size,
            num_outputs=num_outputs,
            constructor_arguments=constructor_args,
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )

        self._backbone = model

    def get_features(self, *args, **kwargs):
        return self._backbone(*args, **kwargs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Dict, Tuple, Type

import pytest

from conftest import L2P_LEARNERS, LEARNERS, LEARNER_KWARGS
from renate.benchmark.models.l2p import LearningToPromptTransformer
from renate.models import RenateModule
from renate.updaters.learner import Learner


def get_model_and_learner_and_learner_kwargs(
    learner_class: Type[Learner],
) -> Tuple[RenateModule, Learner, Dict[str, Any]]:
    learner_kwargs = LEARNER_KWARGS[learner_class]
    model = pytest.helpers.get_renate_module_mlp(
        num_inputs=1, num_outputs=1, hidden_size=1, num_hidden_layers=1
    )
    learner = learner_class(
        model=model,
        loss_fn=pytest.helpers.get_loss_fn(),
        optimizer=pytest.helpers.get_partial_optimizer(),
        **learner_kwargs,
    )
    return model, learner, learner_kwargs


def check_learner_variables(learner: Learner, expected_variable_values: Dict[str, Any]):
    for attribute_name, attribute_value in expected_variable_values.items():
        if attribute_name in [
            "memory_size",
            "learner_class_name",
            "val_memory_buffer",
            "memory_buffer",
        ]:
            continue
        assert getattr(learner, f"_{attribute_name}") == attribute_value


@pytest.mark.parametrize("learner_class", LEARNERS)
def test_save_and_load_learner(learner_class):
    if learner_class not in L2P_LEARNERS:
        model, learner, learner_kwargs = get_model_and_learner_and_learner_kwargs(learner_class)
        checkpoint_dict = {}
        learner.on_save_checkpoint(checkpoint=checkpoint_dict)
        check_learner_variables(learner, checkpoint_dict)
    else:
        with pytest.raises(AssertionError):
            model, learner, learner_kwargs = get_model_and_learner_and_learner_kwargs(learner_class)

        learner_kwargs = LEARNER_KWARGS[learner_class]
        model = LearningToPromptTransformer()
        learner = learner_class(
            model=model,
            loss_fn=pytest.helpers.get_loss_fn(),
            optimizer=pytest.helpers.get_partial_optimizer(),
            **learner_kwargs,
        )
        checkpoint_dict = {}
        learner.on_save_checkpoint(checkpoint=checkpoint_dict)
        check_learner_variables(learner, checkpoint_dict)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from collections import Counter

import numpy as np
import pytest
import torch
from torchvision.transforms.functional import rotate

from dummy_datasets import DummyDataIncrementalDataModule, DummyTorchVisionDataModule
from renate.benchmark.datasets.vision_datasets import TorchVisionDataModule
from renate.benchmark.scenarios import (
    ClassIncrementalScenario,
    DataIncrementalScenario,
    FeatureSortingScenario,
    IIDScenario,
    ImageRotationScenario,
    PermutationScenario,
)
from renate.utils.pytorch import randomly_split_data


@pytest.mark.parametrize(
    "scenario_cls,kwargs",
    [
        [
            ImageRotationScenario,  # Wrong chunk id
            {"num_tasks": 3, "chunk_id": 4, "degrees": [0, 60, 180]},
        ],
        [
            ClassIncrementalScenario,
            {
                "num_tasks": 3,
                "chunk_id": 4,  # Wrong chunk id
                "groupings": ((0, 1, 2), (3, 5, 9), (4, 6, 7, 8)),
            },
        ],
        [PermutationScenario, {"num_tasks": 3, "chunk_id": 2}],  # Missing input dim
    ],
)
def test_failing_to_init(tmpdir, scenario_cls, kwargs):
    dataset_name = "FashionMNIST"
    data_module = TorchVisionDataModule(
        tmpdir, src_bucket=None, src_object_name=None, dataset_name=dataset_name
    )
    with pytest.raises(Exception):
        scenario_cls(data_module=data_module, **kwargs)


def test_class_incremental_scenario():
    data_module = DummyTorchVisionDataModule(val_size=0.3, seed=42)
    groupings = ((0, 1, 3), (2,), (3, 4))
    train_data_class_counts = Counter({3: 16, 4: 15, 0: 15, 2: 13, 1: 11})
    val_data_class_counts = Counter({1: 9, 2: 7, 4: 5, 0: 5, 3: 4})
    test_data_class_counts = Counter({0: 20, 1: 20, 2: 20, 3: 20, 4: 20})
    for i in range(len(groupings)):
        scenario = ClassIncrementalScenario(
            data_module=data_module, groupings=groupings, chunk_id=i
        )
        scenario.prepare_data()
        scenario.setup()
        train_data = scenario.train_data()
        val_data = scenario.val_data()
        assert len(train_data) == sum([train_data_class_counts[c] for c in groupings[i]])
        assert len(val_data) == sum([val_data_class_counts[c] for c in groupings[i]])
        for j, test_data in enumerate(scenario.test_data()):
            assert len(test_data) == sum([test_data_class_counts[c] for c in groupings[j]])


def test_class_incremental_scenario_groupings_error():
    """Classes selected do not exist in data."""
    scenario = ClassIncrementalScenario(
        data_module=DummyTorchVisionDataModule(val_size=0.3, seed=42),
        groupings=((0, 1, 3), (2, 200)),
        chunk_id=0,
    )
    scenario.prepare_data()
    with pytest.raises(ValueError, match=r"Chunk 1 does not contain classes \[200\]."):
        scenario.setup()


def test_data_incremental_scenario_grouping():
    """Test whether grouping in DataIncrementalScenario works"""
    scenario = DataIncrementalScenario(
        data_module=DummyDataIncrementalDataModule(0, (10, 1), val_size=0.2),
        chunk_id=0,
        groupings=((0, 1), (2,)),
    )
    scenario.prepare_data()
    scenario.setup()
    assert set(int(scenario.train_data()[i][1]) for i in range(80)) == {0}
    assert set(int(scenario.train_data()[i][1]) for i in range(80, 160)) == {1}
    assert set(int(scenario.val_data()[i][1]) for i in range(20)) == {0}
    assert set(int(scenario.val_data()[i][1]) for i in range(20, 40)) == {1}
    assert set(int(scenario.test_data()[0][i][1]) for i in range(100)) == {0}
    assert set(int(scenario.test_data()[0][i][1]) for i in range(100, 200)) == {1}
    assert set(int(scenario.test_data()[1][i][1]) for i in range(100)) == {2}
    assert len(scenario.train_data()) == 160
    assert len(scenario.val_data()) == 40
    assert len(scenario.test_data()) == 2
    assert len(scenario.test_data()[0]) == 200
    assert len(scenario.test_data()[1]) == 100


def test_data_incremental_scenario_data_ids():
    """Test whether data_ids in DataIncrementalScenario works"""
    scenario = DataIncrementalScenario(
        data_module=DummyDataIncrementalDataModule(0, (10, 1), val_size=0.2),
        chunk_id=1,
        data_ids=(3, 4),
    )
    scenario.prepare_data()
    scenario.setup()
    assert set(int(scenario.train_data()[i][1]) for i in range(80)) == {4}
    assert set(int(scenario.val_data()[i][1]) for i in range(20)) == {4}
    assert set(int(scenario.test_data()[0][i][1]) for i in range(100)) == {3}
    assert set(int(scenario.test_data()[1][i][1]) for i in range(100)) == {4}
    assert len(scenario.train_data()) == 80
    assert len(scenario.val_data()) == 20
    assert len(scenario.test_data()) == 2
    assert len(scenario.test_data()[0]) == 100
    assert len(scenario.test_data()[1]) == 100


def test_data_incremental_scenario_data_module_error():
    """Check that DataIncrementalScenario raises Exception for unsupported DataModule."""
    with pytest.raises(ValueError, match=r"This scenario is only compatible with*"):
        DataIncrementalScenario(
            data_module=DummyTorchVisionDataModule(),
            data_ids=(0, 1),
            chunk_id=0,
        )


@pytest.mark.parametrize(
    "data_ids,groupings,expected_error_message",
    (
        (
            (0, 1),
            ((0,), (1,)),
            "Either `data_ids` or `groupings` must be provided. Both were provided.",
        ),
        (None, None, "Either `data_ids` or `groupings` must be provided. None was provided."),
    ),
)
def test_data_incremental_scenario_wrong_input(data_ids, groupings, expected_error_message):
    """Scenario expect either data_ids or groupings."""
    with pytest.raises(ValueError, match=expected_error_message):
        DataIncrementalScenario(
            data_module=DummyDataIncrementalDataModule(data_id=0, input_shape=(2, 2)),
            data_ids=data_ids,
            groupings=groupings,
            chunk_id=0,
        )


def test_image_rotation_scenario():
    data_module = DummyTorchVisionDataModule(val_size=0.3)
    degrees = [15, 75]
    for i in range(len(degrees)):
        scenario = ImageRotationScenario(
            data_module=data_module,
            degrees=degrees,
            chunk_id=i,
            seed=data_module._seed,
        )
        scenario.prepare_data()
        scenario.setup()
        for stage in ["train", "val"]:
            scenario_data = getattr(scenario, f"{stage}_data")()
            orig_data_module_data = getattr(data_module, f"{stage}_data")()
            split_orig_data_module_data = randomly_split_data(
                orig_data_module_data, [0.5, 0.5], seed=data_module._seed
            )[i]
            assert len(scenario_data) == len(split_orig_data_module_data)
            for j in range(len(scenario_data)):
                assert torch.equal(
                    rotate(split_orig_data_module_data[j][0], degrees[i]), scenario_data[j][0]
                )
        for j, test_data in enumerate(scenario.test_data()):
            assert len(test_data) == len(data_module.test_data())
            for k in range(len(test_data)):
                assert torch.equal(
                    rotate(data_module.test_data()[k][0], degrees[j]), test_data[k][0]
                )


def test_permutation_scenario():
    data_module = DummyTorchVisionDataModule(val_size=0.3)
    for i in range(3):
        scenario = PermutationScenario(
            data_module=data_module,
            num_tasks=3,
            input_dim=np.prod(data_module.input_shape),
            chunk_id=i,
            seed=data_module._seed,
        )
        scenario.prepare_data()
        scenario.setup()
        for stage in ["train", "val"]:
            scenario_data = getattr(scenario, f"{stage}_data")()
            orig_data_module_data = getattr(data_module, f"{stage}_data")()
            split_orig_data_module_data = randomly_split_data(
                orig_data_module_data, [1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0], seed=data_module._seed
            )[i]
            assert len(scenario_data) == len(split_orig_data_module_data)
            for j in range(len(scenario_data)):
                a, _ = torch.sort(split_orig_data_module_data[j][0].flatten())
                b, _ = torch.sort(scenario_data[j][0].flatten())
                assert torch.equal(a, b)
        for j, test_data in enumerate(scenario.test_data()):
            assert len(test_data) == len(data_module.test_data())
            for k in range(len(test_data)):
                a, _ = torch.sort(data_module.test_data()[k][0].flatten())
                b, _ = torch.sort(test_data[k][0].flatten())
                assert torch.equal(a, b)


@pytest.mark.parametrize(
    "scenario_class, scenario_kwargs",
    (
        (ImageRotationScenario, {"degrees": [0, 90, 180, 270]}),
        (PermutationScenario, {"num_tasks": 4, "input_dim": (1, 5, 5)}),
    ),
)
def test_transforms_in_transform_scenarios_are_distinct(scenario_class, scenario_kwargs):
    """Tests if transformations are different.

    Checks two cases: 1) transforms must not be same objects and 2) transforms must not perform
    identical operations.
    """
    data_module = DummyTorchVisionDataModule()
    scenario = scenario_class(data_module=data_module, **scenario_kwargs, chunk_id=0, seed=0)
    scenario.prepare_data()
    scenario.setup()
    x = data_module.X_train[0]
    for i, transform in enumerate(scenario._transforms):
        for j, transform2 in enumerate(scenario._transforms):
            if i == j:
                continue
            assert transform != transform2
            assert not torch.all(torch.isclose(transform(x), transform2(x)))


def test_iid_scenario():
    """Tests that the IID scenario creates a non-overlapping split."""
    data_module = DummyTorchVisionDataModule(val_size=0.3)
    counter = {}
    for i in range(3):
        scenario = IIDScenario(
            data_module=data_module,
            num_tasks=3,
            chunk_id=i,
            seed=data_module._seed,
        )
        scenario.prepare_data()
        scenario.setup()
        for stage in ["train", "val"]:
            scenario_data = getattr(scenario, f"{stage}_data")()
            for j in range(len(scenario_data)):
                x, y = scenario_data[j]
                assert x not in counter
                counter[x] = 1
        assert len(scenario.test_data()) == 3


@pytest.mark.parametrize("feature_idx", (0, 1))
def test_feature_sorting_scenario(feature_idx):
    """Tests the FeatureSortingScenario.

    Checks for increasing feature values across chunks.
    """
    data_module = DummyTorchVisionDataModule(val_size=0.3)
    num_tasks = 3
    max_value = {"train": -float("inf"), "val": -float("inf")}
    for i in range(3):
        scenario = FeatureSortingScenario(
            data_module=data_module,
            num_tasks=num_tasks,
            feature_idx=feature_idx,
            randomness=0,
            chunk_id=i,
            seed=data_module._seed,
        )
        scenario.prepare_data()
        scenario.setup()
        for stage in ["train", "val"]:
            scenario_data = getattr(scenario, f"{stage}_data")()
            features = [x[0, feature_idx].mean().item() for x, _ in scenario_data]
            chunk_min = min(features)
            chunk_max = max(features)
            assert max_value[stage] <= chunk_min
            max_value[stage] = chunk_max
        assert len(scenario.test_data()) == num_tasks



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, List

import torch
from avalanche.core import BasePlugin, SupervisedPlugin
from avalanche.models import NCMClassifier, TrainEvalModel
from avalanche.training import ICaRL, ICaRLLossPlugin
from avalanche.training.plugins import EWCPlugin, LwFPlugin, ReplayPlugin
from avalanche.training.plugins.evaluation import default_evaluator
from avalanche.training.supervised.icarl import _ICaRLPlugin
from avalanche.training.templates import BaseSGDTemplate, SupervisedTemplate
from torch.optim import Optimizer

from renate.updaters.avalanche.plugins import RenateCheckpointPlugin
from renate.updaters.learner import Learner, ReplayLearner
from renate.utils.avalanche import plugin_by_class, remove_plugin, replace_plugin


class AvalancheLoaderMixin:
    """Mixin for Avalanche dummy learner classes."""

    def update_settings(
        self,
        avalanche_learner: BaseSGDTemplate,
        plugins: List[BasePlugin],
        optimizer: Optimizer,
        max_epochs: int,
        device: torch.device,
        eval_every: int,
    ) -> None:
        """Updates settings of Avalanche learner after reloading."""
        avalanche_learner.plugins = remove_plugin(RenateCheckpointPlugin, avalanche_learner.plugins)
        for plugin in plugins:
            avalanche_learner.plugins = replace_plugin(plugin, avalanche_learner.plugins)
        avalanche_learner.model = self._model
        avalanche_learner.optimizer = optimizer
        avalanche_learner._criterion = self._loss_fn
        avalanche_learner.train_epochs = max_epochs
        avalanche_learner.train_mb_size = self._batch_size
        avalanche_learner.eval_mb_size = self._batch_size + getattr(self, "_memory_batch_size", 0)
        avalanche_learner.device = device
        avalanche_learner.eval_every = eval_every

    def _create_avalanche_learner(
        self,
        optimizer: Optimizer,
        train_epochs: int,
        plugins: List[SupervisedPlugin],
        device: torch.device,
        eval_every: int,
        **kwargs: Any,
    ) -> BaseSGDTemplate:
        """Returns Avalanche object that this dummy learner wraps around."""
        return SupervisedTemplate(
            model=self._model,
            optimizer=optimizer,
            criterion=self._loss_fn,
            train_mb_size=self._batch_size,
            eval_mb_size=self._batch_size + getattr(self, "_memory_batch_size", 0),
            train_epochs=train_epochs,
            plugins=plugins,
            evaluator=default_evaluator(),
            device=device,
            eval_every=eval_every,
            **kwargs,
        )


class AvalancheReplayLearner(ReplayLearner, AvalancheLoaderMixin):
    """Renate wrapper around Avalanche Experience Replay."""

    def create_avalanche_learner(
        self, plugins: List[SupervisedPlugin], **kwargs: Any
    ) -> BaseSGDTemplate:
        replay_plugin = ReplayPlugin(
            mem_size=self._memory_buffer._max_size,
            batch_size=self._batch_size,
            batch_size_mem=self._memory_batch_size,
        )
        plugins.append(replay_plugin)
        return self._create_avalanche_learner(plugins=plugins, **kwargs)


class AvalancheEWCLearner(Learner, AvalancheLoaderMixin):
    """Renate wrapper around Avalanche EWC."""

    def __init__(self, ewc_lambda: float, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._ewc_lambda = ewc_lambda

    def update_settings(self, avalanche_learner: BaseSGDTemplate, **kwargs: Any):
        super().update_settings(avalanche_learner=avalanche_learner, **kwargs)
        plugin_by_class(EWCPlugin, avalanche_learner.plugins).ewc_lambda = self._ewc_lambda

    def create_avalanche_learner(
        self, plugins: List[SupervisedPlugin], **kwargs
    ) -> BaseSGDTemplate:
        ewc_plugin = EWCPlugin(ewc_lambda=self._ewc_lambda)
        plugins.append(ewc_plugin)
        return self._create_avalanche_learner(plugins=plugins, **kwargs)


class AvalancheLwFLearner(Learner, AvalancheLoaderMixin):
    """Renate wrapper around Avalanche LwF"""

    def __init__(self, alpha: float, temperature: float, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._alpha = alpha
        self._temperature = temperature

    def update_settings(self, avalanche_learner: BaseSGDTemplate, **kwargs: Any):
        super().update_settings(avalanche_learner=avalanche_learner, **kwargs)
        lwf_plugin = plugin_by_class(LwFPlugin, avalanche_learner.plugins)
        lwf_plugin.lwf.alpha = self._alpha
        lwf_plugin.lwf.temperature = self._temperature

    def create_avalanche_learner(
        self, plugins: List[SupervisedPlugin], **kwargs
    ) -> BaseSGDTemplate:
        lwf_plugin = LwFPlugin(alpha=self._alpha, temperature=self._temperature)
        plugins.append(lwf_plugin)
        return self._create_avalanche_learner(plugins=plugins, **kwargs)


class AvalancheICaRLLearner(Learner, AvalancheLoaderMixin):
    """Renate wrapper around Avalanche ICaRL."""

    def __init__(self, memory_size: int, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._memory_size = memory_size

    def create_avalanche_learner(
        self,
        optimizer: Optimizer,
        train_epochs: int,
        plugins: List[SupervisedPlugin],
        device: torch.device,
        eval_every: int,
    ) -> BaseSGDTemplate:
        if not hasattr(self._model, "class_means"):
            raise RuntimeError(
                """The RenateModule must contain an attribute `class_means`.
                Please add something like

                self.class_means = torch.nn.Parameter(
                    torch.zeros((embedding_size, num_outputs)), requires_grad=False
                )
                """
            )
        if not hasattr(self._model, "get_backbone") or not hasattr(self._model, "get_predictor"):
            raise RuntimeError(
                "The RenateModule must be explicitly split into backbone and predictor module. "
                "Please implement functions `get_backbone()` and `get_predictor()` to return these "
                "modules."
            )
        icarl = ICaRL(
            feature_extractor=self._model.get_backbone(),
            classifier=self._model.get_predictor(),
            optimizer=optimizer,
            memory_size=self._memory_size,
            buffer_transform=None,  # TODO
            fixed_memory=True,
            train_mb_size=self._batch_size,
            train_epochs=train_epochs,
            eval_mb_size=self._batch_size,
            device=device,
            plugins=plugins,
            eval_every=-1,  # TODO: https://github.com/ContinualAI/avalanche/issues/1281
        )
        plugin_by_class(_ICaRLPlugin, icarl.plugins).class_means = self._model.class_means

        return icarl

    def update_settings(self, avalanche_learner: BaseSGDTemplate, **kwargs) -> None:
        super().update_settings(avalanche_learner=avalanche_learner, **kwargs)
        avalanche_learner.model = TrainEvalModel(
            feature_extractor=self._model.get_backbone(),
            train_classifier=self._model.get_predictor(),
            eval_classifier=NCMClassifier(),
        )
        icarl_loss_plugin = plugin_by_class(ICaRLLossPlugin, avalanche_learner.plugins)
        avalanche_learner._criterion = icarl_loss_plugin
        avalanche_learner.eval_every = (
            -1
        )  # TODO: https://github.com/ContinualAI/avalanche/issues/1281



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import warnings
from copy import deepcopy

import pytest
import torch
from pytorch_lightning.utilities.seed import seed_everything
from torchvision.transforms import Lambda

from conftest import LEARNERS_USING_SIMPLE_UPDATER, LEARNER_KWARGS, check_learner_transforms
from renate import defaults
from renate.updaters.experimental.repeated_distill import RepeatedDistillationModelUpdater
from renate.updaters.learner import ReplayLearner


@pytest.mark.parametrize("provide_folder", [True, False])
def test_simple_model_updater(tmpdir, provide_folder):
    model, train_dataset, test_data = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    model_updater = pytest.helpers.get_simple_updater(
        model, output_state_folder=defaults.output_state_folder(tmpdir) if provide_folder else None
    )
    y_hat_before_train = model(test_data, task_id=defaults.TASK_ID)
    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    y_hat_after_train = model(test_data, task_id=defaults.TASK_ID)
    assert y_hat_before_train.shape[0] == y_hat_after_train.shape[0]
    assert not torch.allclose(y_hat_before_train, y_hat_after_train)


def test_model_passed_is_used_as_is(tmpdir):
    """Makes sure that the model passed to the updater is not overwritten by anything in the
    checkpoint"""
    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    model2 = deepcopy(model)
    model_updater = pytest.helpers.get_simple_updater(model, output_state_folder=tmpdir)
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)

    expected_model = deepcopy(model2)
    model_updater = pytest.helpers.get_simple_updater(model2, input_state_folder=tmpdir)
    for p1, p2 in zip(
        expected_model.parameters(),
        model_updater._learner._model.parameters(),
    ):
        assert torch.allclose(p1, p2)


def test_deterministic_updater():
    # The behavior is always deterministic on CPU, but it can become non-deterministic on GPU
    # When run on CPU this test never fails, so it is only useful when tests are run on GPU
    model1, train_dataset, test_data = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )

    model2 = deepcopy(model1)

    model_updater1 = pytest.helpers.get_simple_updater(
        model1,
        deterministic_trainer=True,
    )

    model_updater2 = pytest.helpers.get_simple_updater(
        model2,
        deterministic_trainer=True,
    )

    model1 = model_updater1.update(train_dataset, task_id=defaults.TASK_ID)
    model2 = model_updater2.update(train_dataset, task_id=defaults.TASK_ID)

    y_hat_1 = model1(test_data, task_id=defaults.TASK_ID)
    y_hat_2 = model2(test_data, task_id=defaults.TASK_ID)

    assert torch.allclose(y_hat_1, y_hat_2)


@pytest.mark.parametrize("early_stopping_enabled", [True, False])
@pytest.mark.parametrize("use_val", [True, False])
@pytest.mark.parametrize("metric_monitored", [None, "val_loss"])
@pytest.mark.parametrize("updater_type", ["DMC", "SimpleUpdater"])
def test_model_updater_with_early_stopping(
    use_val, early_stopping_enabled, metric_monitored, updater_type
):
    seed_everything(0)
    model, train_dataset, val_dataset = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=8,
        num_hidden_layers=1,
        train_num_samples=10,
        test_num_samples=5,
        val_num_samples=10,
    )

    max_epochs = 10
    with warnings.catch_warnings(record=True) as warning_init:
        if updater_type == "DMC":
            model_updater = RepeatedDistillationModelUpdater(
                model=model,
                loss_fn=pytest.helpers.get_loss_fn(),
                optimizer=pytest.helpers.get_partial_optimizer(lr=0.3),
                memory_size=50,
                max_epochs=max_epochs,
                early_stopping_enabled=early_stopping_enabled,
                metric=metric_monitored,
                accelerator="cpu",
                devices=1,
            )
        else:
            model_updater = pytest.helpers.get_simple_updater(
                model=model,
                max_epochs=max_epochs,
                early_stopping_enabled=early_stopping_enabled,
                metric=metric_monitored,
            )

    assert model_updater._num_epochs_trained == 0
    with warnings.catch_warnings(record=True) as warning_update:
        model_updater.update(
            train_dataset, val_dataset=val_dataset if use_val else None, task_id=defaults.TASK_ID
        )

    if metric_monitored and use_val and early_stopping_enabled:
        assert model_updater._num_epochs_trained < max_epochs
    else:
        assert model_updater._num_epochs_trained == max_epochs

    is_warning_metric_missing_sent = any(
        [
            str(w.message).startswith("Early stopping is enabled but no metric is specified")
            for w in warning_init
        ]
    )
    is_warning_early_stopping_without_val_set_sent = any(
        [
            str(w.message).startswith(
                "Early stopping is currently not supported without a validation set"
            )
            for w in warning_update
        ]
    )
    assert not metric_monitored and early_stopping_enabled or not is_warning_metric_missing_sent
    assert (
        metric_monitored
        and early_stopping_enabled
        and not use_val
        or not is_warning_early_stopping_without_val_set_sent
    )


@pytest.mark.parametrize("learner_class", LEARNERS_USING_SIMPLE_UPDATER)
def test_continuation_of_training_with_simple_model_updater(tmpdir, learner_class):
    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=1,
        train_num_samples=10,
        test_num_samples=5,
    )
    state_url = defaults.input_state_folder(tmpdir)
    model_updater = pytest.helpers.get_simple_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=LEARNER_KWARGS[learner_class],
        output_state_folder=state_url,
        max_epochs=2,
    )
    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    model_updater = pytest.helpers.get_simple_updater(
        model,
        learner_class=learner_class,
        learner_kwargs=LEARNER_KWARGS[learner_class],
        input_state_folder=state_url,
        max_epochs=2,
    )
    model_updater.update(train_dataset, task_id=defaults.TASK_ID)


@pytest.mark.parametrize("learner_class", LEARNERS_USING_SIMPLE_UPDATER)
def test_transforms_passed_to_simple_model_updater_will_be_used_by_learner(tmpdir, learner_class):
    """Checks if all transforms are correctly forwarded to Learner and MemoryBuffer.

    There are two cases: passing transforms
        1) when creating Learner from scratch and
        2) when loading Learner from checkpoint
    """

    def identity_transform():
        return Lambda(lambda x: x)

    model, train_dataset, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    state_url = defaults.input_state_folder(tmpdir)
    transforms_kwargs = {
        "train_transform": identity_transform(),
        "train_target_transform": identity_transform(),
        "test_transform": identity_transform(),
        "test_target_transform": identity_transform(),
    }
    if issubclass(learner_class, ReplayLearner):
        transforms_kwargs["buffer_transform"] = identity_transform()
        transforms_kwargs["buffer_target_transform"] = identity_transform()
    model_updater = pytest.helpers.get_simple_updater(
        model,
        output_state_folder=state_url,
        learner_kwargs=LEARNER_KWARGS[learner_class],
        learner_class=learner_class,
        **transforms_kwargs,
    )
    check_learner_transforms(model_updater._learner, transforms_kwargs)
    model = model_updater.update(train_dataset, task_id=defaults.TASK_ID)
    model_updater = pytest.helpers.get_simple_updater(
        model,
        input_state_folder=state_url,
        learner_class=learner_class,
        **transforms_kwargs,
        learner_kwargs=LEARNER_KWARGS[learner_class],
    )
    check_learner_transforms(model_updater._learner, transforms_kwargs)


@pytest.mark.xfail(raises=(KeyError, TypeError))
@pytest.mark.parametrize("learner", LEARNERS_USING_SIMPLE_UPDATER)
def test_learner_fails_without_loss_fn(learner):
    """This test checks that the updater crashes when it is supposed to.
    This is to check for loss_fn and other misc arguments."""
    _ = pytest.helpers.get_simple_updater(
        model=torch.nn.Linear(1, 1),
        learner_class=learner,
        learner_kwargs={"learning_rate": 0.0},
        max_epochs=1,
    )
